# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
Definitions for compossible types to run the SHAKE permute cipher
mode of operation.
"""

__all__ = [
    "ShakePermuteKDFs",
    "ShakePermuteSessionKDFs",
    "ShakePermuteKeyAADBundle",
    "ShakePermuteStreamHMAC",
    "ShakePermuteStreamJunction",
]


from aiootp._typing import Typing as t
from aiootp._constants.misc import BIG
from aiootp._exceptions import Issue
from aiootp.asynchs import asleep
from aiootp.commons import FrozenInstance

from .cipher_kdfs import CipherKDFs
from .key_bundle import KeyAADBundle
from .stream_hmac import StreamHMAC
from .stream_junction import StreamJunction
from .slick_256_config import SHMAC_KDF


class ShakePermuteKDFs(CipherKDFs):
    """
    A type responsible for initializing the key-derivation & mac objects
    for SHAKE permute mode ciphers.
    """

    __slots__ = ("keyed_shmac_kdf",)

    def __init__(self, key: bytes, *, config: t.ConfigType) -> None:
        """
        Keys the KDFs for handling by `KeyAADBundle`.
        """
        self.config = config
        self.test_key_validity(key)
        self.keyed_shmac_kdf = self.key_base_kdf(SHMAC_KDF, key=key)

    def new_session(
        self, summary: bytes
    ) -> t.Generator[t.Tuple[str, t.XOFType], None, None]:
        """
        Yields copies of the instance's KDFs that have been given the
        fresh randomness & context of a session's `summary`.
        """
        yield SHMAC_KDF, self.keyed_shmac_kdf.new_session_copy(summary)


class ShakePermuteSessionKDFs(FrozenInstance):
    """
    Efficiently stores & gives access to keyed & session-randomized /
    contextualized functions.
    """

    __slots__ = ("shmac_kdf",)

    def __init__(self) -> None:
        pass


class ShakePermuteKeyAADBundle(KeyAADBundle):
    """
    A low-level interface for managing a key, salt, iv & authenticated
    associated data bundle which is to be used for ONLY ONE encryption.
    """

    __slots__ = ()

    _Session: type = ShakePermuteSessionKDFs

    async def async_mode(self) -> t.Self:
        """
        Sets the instance to run async SHAKE permute mode key derivation.
        """
        await asleep()
        self._mode.set_async_mode()
        return self

    def sync_mode(self) -> t.Self:
        """
        Sets the instance to run sync SHAKE permute mode key derivation.
        """
        self._mode.set_sync_mode()
        return self


class ShakePermuteStreamHMAC(StreamHMAC):
    """
    This class is used as an inline validator for ciphertext streams as
    they are being created & decrypted. It's designed for full context
    commitment, salt misuse-reuse resistance, & RUP security.
    """

    __slots__ = ()

    async def _aencipher_then_hash(
        self,
        plaintext: bytes,
        *,
        _from_bytes: t.Callable[[bytes, str], int] = int.from_bytes,
    ) -> bytes:
        """
        This method is inserted as the instance's `_avalidated_transform`
        method after the encryption mode is set.
        """
        # fmt: off
        c = self.config
        try:
            digest = self._current_digest[c.PERMUTATION_DIGEST_SLICE]
            in_key = _from_bytes(self._current_digest[c.IN_KEY_SLICE], BIG)
            out_key = _from_bytes(self._current_digest[c.OUT_KEY_SLICE], BIG)
            ciphertext = (
                out_key
                ^ await self._permutation.auncapped_permute(
                    in_key ^ _from_bytes(plaintext, BIG)
                )
            ).to_bytes(c.BLOCKSIZE, BIG)
            await self._aupdate(plaintext + ciphertext + digest)  # 168 bytes
            return ciphertext
        except OverflowError as error:                              # pragma: no cover
            raise Issue.exceeded_blocksize(c.BLOCKSIZE) from error  # pragma: no cover
        # fmt: on

    def _encipher_then_hash(
        self,
        plaintext: bytes,
        *,
        _from_bytes: t.Callable[[bytes, str], int] = int.from_bytes,
    ) -> bytes:
        """
        This method is inserted as the instance's `_validated_transform`
        method after the encryption mode is set.
        """
        # fmt: off
        c = self.config
        try:
            digest = self._current_digest[c.PERMUTATION_DIGEST_SLICE]
            in_key = _from_bytes(self._current_digest[c.IN_KEY_SLICE], BIG)
            out_key = _from_bytes(self._current_digest[c.OUT_KEY_SLICE], BIG)
            ciphertext = (
                out_key
                ^ self._permutation.uncapped_permute(
                    in_key ^ _from_bytes(plaintext, BIG)
                )
            ).to_bytes(c.BLOCKSIZE, BIG)
            self._update(plaintext + ciphertext + digest)  # 168 bytes
            return ciphertext
        except OverflowError as error:                              # pragma: no cover
            raise Issue.exceeded_blocksize(c.BLOCKSIZE) from error  # pragma: no cover
        # fmt: on

    async def _ahash_then_decipher(
        self,
        ciphertext: bytes,
        *,
        _from_bytes: t.Callable[[bytes, str], int] = int.from_bytes,
    ) -> bytes:
        """
        This method is inserted as the instance's `_avalidated_transform`
        method after the decryption mode is set.
        """
        # fmt: off
        c = self.config
        try:
            digest = self._current_digest[c.PERMUTATION_DIGEST_SLICE]
            in_key = _from_bytes(self._current_digest[c.IN_KEY_SLICE], BIG)
            out_key = _from_bytes(self._current_digest[c.OUT_KEY_SLICE], BIG)
            plaintext = (
                in_key
                ^ await self._permutation.auncapped_invert(
                    out_key ^ _from_bytes(ciphertext, BIG)
                )
            ).to_bytes(c.BLOCKSIZE, BIG)
            await self._aupdate(plaintext + ciphertext + digest)  # 168 bytes
            return plaintext
        except OverflowError as error:                              # pragma: no cover
            raise Issue.exceeded_blocksize(c.BLOCKSIZE) from error  # pragma: no cover
        # fmt: on

    def _hash_then_decipher(
        self,
        ciphertext: bytes,
        *,
        _from_bytes: t.Callable[[bytes, str], int] = int.from_bytes,
    ) -> bytes:
        """
        This method is inserted as the instance's `_avalidated_transform`
        method after the decryption mode is set.
        """
        # fmt: off
        c = self.config
        try:
            digest = self._current_digest[c.PERMUTATION_DIGEST_SLICE]
            in_key = _from_bytes(self._current_digest[c.IN_KEY_SLICE], BIG)
            out_key = _from_bytes(self._current_digest[c.OUT_KEY_SLICE], BIG)
            plaintext = (
                in_key
                ^ self._permutation.uncapped_invert(
                    out_key ^ _from_bytes(ciphertext, BIG)
                )
            ).to_bytes(c.BLOCKSIZE, BIG)
            self._update(plaintext + ciphertext + digest)  # 168 bytes
            return plaintext
        except OverflowError as error:                              # pragma: no cover
            raise Issue.exceeded_blocksize(c.BLOCKSIZE) from error  # pragma: no cover
        # fmt: on


class ShakePermuteStreamJunction(StreamJunction):
    """
    A definition for how the key & data streams are combined for the
    SHAKE permute cipher mode.
    """

    @classmethod
    async def acombine_streams(
        cls, data: t.AsyncDatastream, *, shmac: StreamHMAC
    ) -> t.AsyncGenerator[bytes, None]:
        """
        XORs each ciphertext or plaintext block with a SHAKE round key,
        then passes the sum through a keyed permutation before XORing
        with another SHAKE round key, & feeding the values to the `shmac`
        for validation & distinct altering of the keystream.
        """
        datastream, validated_transform = (
            data,
            shmac._avalidated_transform,
        )
        async for block in datastream:
            yield await validated_transform(block)

    @classmethod
    def combine_streams(
        cls, data: t.Datastream, *, shmac: StreamHMAC
    ) -> t.Generator[bytes, None, None]:
        """
        XORs each ciphertext or plaintext block with a SHAKE round key,
        then passes the sum through a keyed permutation before XORing
        with another SHAKE round key, & feeding the values to the `shmac`
        for validation & distinct altering of the keystream.
        """
        datastream, validated_transform = (data, shmac._validated_transform)
        for block in datastream:
            yield validated_transform(block)


module_api = dict(
    ShakePermuteKDFs=t.add_type(ShakePermuteKDFs),
    ShakePermuteSessionKDFs=t.add_type(ShakePermuteSessionKDFs),
    ShakePermuteKeyAADBundle=t.add_type(ShakePermuteKeyAADBundle),
    ShakePermuteStreamHMAC=t.add_type(ShakePermuteStreamHMAC),
    ShakePermuteStreamJunction=t.add_type(ShakePermuteStreamJunction),
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
)



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
A type for converting secret `int`s into bytes-type identification
numbers using a bijective, keyed permutation.
"""

__all__ = ["SequenceID"]


from aiootp._typing import Typing as t
from aiootp._constants import BIG
from aiootp._permutations import FastAffineXORChain
from aiootp.commons import ConfigMap, FrozenInstance

from .sequence_id_config import SequenceIDConfig


class SequenceID(FrozenInstance):
    """
    A class for producing unique, deterministic pseudo-random sequential
    identifiers from integer indexes. The identifiers don't suffer from
    birthday collisions if the difference between the minimum & maximum
    index values is no larger than 256**size, where 'size' is the number
    of bytes of the produced identifiers.

    The produced identifiers are randomized by a unique, secret, uniform
    key. Any instance created using the key is able to create the same
    sequence of identifiers. Normal birthday collision probabilities
    apply when a different key is used.

     _____________________________________
    |                                     |
    |            Usage Example:           |
    |_____________________________________|

    from aiootp.randoms import SequenceID, token_bytes
    from aiootp.ciphers import ChaCha20Poly1305
    from aiootp.keygens import DomainKDF

    seed = token_bytes(32)
    kdf = DomainKDF(b"movie-server", seed, key=session.shared_key)
    key = kdf.shake_128(size=32 + SequenceID.key_size(12), aad=b"joint-key")
    cipher = ChaCha20Poly1305(key[:32])
    sid = SequenceID(key=key[32:], config_id=12)

    yield seed
    for i, movie in enumerate(movie_collection):
        yield cipher.encrypt(
            nonce=sid.new(i), data=movie, associated_data=session.aad
        )
    """

    __slots__ = ("_permutation", "config")

    _configs = ConfigMap(
        {
            config_id: SequenceIDConfig(
                config_id=config_id,
                size=config_id,
                permutation_type=FastAffineXORChain,
                permutation_config_id=config_id,
            )
            for config_id in range(1, 33)
        },
        config_type=SequenceIDConfig,
    )

    @classmethod
    def key_size(cls, config_id: t.Hashable) -> int:
        """
        Returns the number of bytes a uniform random key needs to be
        to initialize the instance's permutation.
        """
        return cls._configs[config_id].KEY_SIZE

    def _initialize_permutation(self, key: bytes) -> t.PermutationType:
        """
        Returns a bijective, keyed permutation as specified by the
        instance's configuration.
        """
        config_id = self.config.PERMUTATION_CONFIG_ID
        return self.config.Permutation(key=key, config_id=config_id)

    def __init__(
        self, *, key: t.Optional[bytes], config_id: t.Hashable = 12
    ) -> None:
        """
        Initialized the keyed permutation.
        """
        self.config = self._configs[config_id]
        self._permutation = self._initialize_permutation(key)

    async def anew(self, value: int) -> bytes:
        """
        Permutes the secret `value` using a bijective, keyed permutation.
        """
        sid = await self._permutation.apermute(value)
        return sid.to_bytes(self.config.SIZE, BIG)

    def new(self, value: int) -> bytes:
        """
        Permutes the secret `value` using a bijective, keyed permutation.
        """
        sid = self._permutation.permute(value)
        return sid.to_bytes(self.config.SIZE, BIG)

    async def aread(self, value: bytes) -> int:
        """
        Returns the secret int that created the permutation `value`.
        """
        return await self._permutation.ainvert(int.from_bytes(value, BIG))

    def read(self, value: bytes) -> int:
        """
        Returns the secret int that created the permutation `value`.
        """
        return self._permutation.invert(int.from_bytes(value, BIG))


module_api = dict(
    SequenceID=t.add_type(SequenceID),
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
)



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
Cryptographically secure random number generation for the package.
"""

__all__ = [
    "acsprng",
    "arandom_number_generator",
    "csprng",
    "random_number_generator",
]


import io
from collections import deque
from hashlib import shake_256

from aiootp._typing import Typing as t
from aiootp._constants import BIG, SHAKE_256_BLOCKSIZE
from aiootp._exceptions import Issue
from aiootp._paths import update_salt_file
from aiootp.asynchs import asleep, gather, new_event_loop
from aiootp.generics import acanonical_pack

from .simple import atoken_bits, token_bytes, arandom_sleep
from .simple import acanonical_token, canonical_token
from .threading_safe_entropy_pool import ThreadingSafeEntropyPool
from .entropy_daemon import EntropyDaemon
from ._early_salts import _salt, _asalt, _asalt_multiply
from ._early_salts import _package_seed, _package_seed_path


# initialize rudimentary global entropy pool
_pool = deque((token_bytes(32) for _ in range(32)), maxlen=32)


# initialize the global SHA3 hashing object that also collects entropy
_gadget = ThreadingSafeEntropyPool(
    _package_seed + token_bytes(3 * SHAKE_256_BLOCKSIZE),
    obj=shake_256,
    pool=_pool,
)


# begin the entropy gathering daemon
_entropy_daemon = EntropyDaemon(
    gadget=_gadget, entropy_pool=_pool, max_delay=1
).start()
_entropy_daemon.set_temporary_max_delay(max_delay=0.001, duration=2)


# avert event loop clashes
_run = new_event_loop().run_until_complete


async def acsprng(
    size: int = 64,
    *,
    entropy: t.Any = _gadget.hash(
        _salt(pool=_pool, gadget=_gadget).to_bytes(32, BIG), size=16
    ),
) -> bytes:
    """
    Returns a `size`-byte cryptographically secure pseudo-random value.
    Optionally seeds the internal entropy pools with user-provided
    `entropy`.
    """
    if entropy.__class__ is bytes:
        token = await acanonical_token() + entropy
    else:
        token = await acanonical_token() + repr(entropy).encode()
    _gadget.update(token)
    thread_safe_entropy = _gadget._obj.copy()
    thread_safe_entropy.update(token + _pool[0])
    return thread_safe_entropy.digest(size)


def csprng(
    size: int = 64,
    *,
    entropy: t.Any = _gadget.hash(
        _salt(pool=_pool, gadget=_gadget).to_bytes(32, BIG), size=16
    ),
) -> bytes:
    """
    Returns a `size`-byte cryptographically secure pseudo-random value.
    Optionally seeds the internal entropy pools with user-provided
    `entropy`.
    """
    if entropy.__class__ is bytes:
        token = canonical_token() + entropy
    else:
        token = canonical_token() + repr(entropy).encode()
    _gadget.update(token)
    thread_safe_entropy = _gadget._obj.copy()
    thread_safe_entropy.update(token + _pool[0])
    return thread_safe_entropy.digest(size)


async def arandom_number_generator(
    size: int = 64, *, entropy: t.Any = csprng(32), freshness: int = 8
) -> bytes:
    """
    Returns a `size`-byte random bytestring derived from the package's
    entropy pools, the provided `entropy` value, a persistent sha3
    hashing object, & a multithreaded + asynchronous chaotic algorithm.
    The integer `freshness` parameter adjusts the number of concurrent
    asynchronous tasks that will be spawned to gather entropy.

     _____________________________________
    |                                     |
    |        Algorithm Explanation:       |
    |_____________________________________|

    The methods used to implement a non-deterministic CSPRNG are given
    below.

    0. Utilize various component PRNGs whose designs & entropy sources
    differ.

    1. Keep the sizes of seeds, intermediate & output entropy values at
    least as large as the minumum acceptable security strength.

    2. Generate new seeds before & after each use.

    3. Use a persistent `shake_256` object on the input, intermediate &
    output values of component PRNGs for backtracking resistance,
    prediction resistance & as an efficient entropy pool.

    4. Ensure an order prediction attack must be carried out on internal
    calculations by using nanosecond timestamps, non-commutative
    operations, & hashing between & internally to each component PRNG.

    5. Make order-prediction more difficult by running a background
    entropy collection thread to asynchronously gather PRNG tasks in
    pseudo-random order in between pseudo-random sleeps, all the while
    ratcheting internal states.

    6. Save, retrieve & update entropy saved to the filesystem every
    time the package is imported.

    7. Hash userland entropy into the entropy pools.

    8. Incorporate function call context into seeds for thread safety &
    domain separation.

    9. Enable the iteration & interleaving of these methods a variable
    amount of times as needed with a `freshness` parameter.

    If we can assume each initial 32-byte internal state produced by the
    CSPRNG only adds 8-bits of entropy to the pool, then by initializing
    the pool to 32 ratcheting states, & hashing it with the persistent
    `shake_256` object, then 256-bits of entropy will have been reached.
    """
    _entropy_daemon.set_temporary_max_delay(0.001, duration=1)

    if freshness < 0 or freshness.__class__ is not int:
        raise Issue.value_must("freshness", "be a non-negative int")
    elif not freshness:
        _pool.appendleft(await acsprng(32, entropy=entropy))
    else:

        async def create_unique_multiple(seed: int) -> int:
            return await _asalt_multiply(
                size, seed, await atoken_bits(256), pool=_pool
            )

        async def modular_multiplication() -> None:
            seed = await _asalt(pool=_pool, gadget=_gadget)
            await arandom_sleep(0.003)
            multiples = (create_unique_multiple(seed) for _ in range(3))
            multiples = [await multiple for multiple in multiples]
            element = await _asalt_multiply(seed, *multiples, pool=_pool)
            result = seed.to_bytes(32, BIG) + element.to_bytes(32, BIG)
            _pool.appendleft(await acsprng(32, entropy=result))

        async def add_to_pool() -> None:
            seed = await acsprng(32, entropy=entropy)
            await arandom_sleep(0.003)
            _pool.appendleft(await acsprng(32, entropy=seed))

        async def start_generator() -> None:
            tasks = deque()
            for _ in range(freshness):
                await asleep()
                tasks.appendleft(modular_multiplication())
                for _ in range(10):
                    tasks.appendleft(add_to_pool())
            reader = io.BytesIO(await acsprng(16 * len(tasks))).read
            await gather(*sorted(tasks, key=lambda _: reader(16)))

        entropy = await acsprng(16, entropy=entropy)
        await start_generator()

    return await acsprng(size, entropy=await acanonical_pack(*_pool))


def random_number_generator(
    size: int = 64, *, entropy: t.Any = csprng(32), freshness: int = 8
) -> bytes:
    """
    Returns a `size`-byte random bytestring derived from the package's
    entropy pools, the provided `entropy` value, a persistent sha3
    hashing object, & a multithreaded + asynchronous chaotic algorithm.
    The integer `freshness` parameter adjusts the number of concurrent
    asynchronous tasks that will be spawned to gather entropy.

     _____________________________________
    |                                     |
    |        Algorithm Explanation:       |
    |_____________________________________|

    The methods used to implement a non-deterministic CSPRNG are given
    below.

    0. Utilize various component PRNGs whose designs & entropy sources
    differ.

    1. Keep the sizes of seeds, intermediate & output entropy values at
    least as large as the minumum acceptable security strength.

    2. Generate new seeds before & after each use.

    3. Use a persistent `shake_256` object on the input, intermediate &
    output values of component PRNGs for backtracking resistance,
    prediction resistance & as an efficient entropy pool.

    4. Ensure an order prediction attack must be carried out on internal
    calculations by using nanosecond timestamps, non-commutative
    operations, & hashing between & internally to each component PRNG.

    5. Make order-prediction more difficult by running a background
    entropy collection thread to asynchronously gather PRNG tasks in
    pseudo-random order in between pseudo-random sleeps, all the while
    ratcheting internal states.

    6. Save, retrieve & update entropy saved to the filesystem every
    time the package is imported.

    7. Hash userland entropy into the entropy pools.

    8. Incorporate function call context into seeds for thread safety &
    domain separation.

    9. Enable the iteration & interleaving of these methods a variable
    amount of times as needed with a `freshness` parameter.

    If we can assume each initial 32-byte internal state produced by the
    CSPRNG only adds 8-bits of entropy to the pool, then by initializing
    the pool to 32 ratcheting states, & hashing it with the persistent
    `shake_256` object, then 256-bits of entropy will have been reached.
    """
    return _run(
        arandom_number_generator(size, entropy=entropy, freshness=freshness)
    )


# update the device seed with new entropy
update_salt_file(
    path=_package_seed_path,
    salt=_run(arandom_number_generator(32, freshness=1)),
)


module_api = dict(
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
    acsprng=acsprng,
    arandom_number_generator=arandom_number_generator,
    csprng=csprng,
    random_number_generator=random_number_generator,
)



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
General classes that expect subclasses to define attributes in
`__slots__`.
"""

__all__ = ["FrozenInstance"]


from os import linesep as sep

from aiootp._typing import Typing as t
from aiootp._exceptions import Issue


class FrozenInstance:
    """
    A non-mapping base class which enables subclasses to define frozen
    instances with proper behavior & efficient `__slots__` declarations.
    This could help with safety by explicitly discouraging mutable state
    where it isn't explicity imposed.
    """

    __slots__ = ()

    def __init__(
        self, mapping: t.Mapping[str, t.Any] = {}, /, **kw: t.Any
    ) -> None:
        """
        Populates instance attributes with user-defined kwargs which
        have been declared in the class' `__slots__`.
        """
        for name, value in {**dict(mapping), **kw}.items():
            setattr(self, name, value)

    def __repr__(self, /) -> str:
        """
        Pretty displays the instance & its attributes.
        """
        has = lambda name: hasattr(self, name)
        get = lambda name: getattr(self, name).__class__
        body = f",{sep}    ".join(
            f"{name}={f'<set>{get(name)}' if has(name) else '<unset>'}"
            for cls in self.__class__.__mro__
            for name in getattr(cls, "__slots__", ())
            if str(name[0]) != "_"
        )
        start = f"{sep}    " if body else ""
        end = f",{sep}" if body else ""
        return f"{self.__class__.__qualname__}({start}{body}{end})"

    def __setattr__(self, name: str, value: t.Any, /) -> None:
        """
        Denies setting attributes after they have already been set.
        """
        if hasattr(self, name):
            raise Issue.cant_reassign_attribute(name)
        object.__setattr__(self, name, value)

    def __delattr__(self, name: str, /) -> None:
        """
        Denies deleting attributes.
        """
        raise Issue.cant_deassign_attribute(name)


module_api = dict(
    FrozenInstance=t.add_type(FrozenInstance),
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
)



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


from conftest import *


async def arun_function(relay: Ignore) -> bool:
    relay.bus.error = getattr(relay, "error", None)
    return True


def run_function(relay: Ignore) -> bool:
    relay.bus.error = getattr(relay, "error", None)
    return True


async def arun_function_return_false(relay: Ignore) -> bool:
    relay.bus.error = getattr(relay, "error", None)
    return False


def run_function_return_false(relay: Ignore) -> bool:
    relay.bus.error = getattr(relay, "error", None)
    return False


def arun_function_order_checking(*, name: str) -> None:
    async def mutate_mapping(relay: Ignore) -> bool:
        setattr(relay.bus, name, relay.bus.count)
        relay.bus.count += 1
        return True

    return mutate_mapping


def run_function_order_checking(*, name: str) -> None:
    def mutate_mapping(relay: Ignore) -> bool:
        setattr(relay.bus, name, relay.bus.count)
        relay.bus.count += 1
        return True

    return mutate_mapping


class TestIgnore:
    async def test_if_else_runs_when_no_exceptions(self) -> None:
        async with Ignore(
            ZeroDivisionError, if_else=arun_function
        ) as relay:
            pass
        assert relay.bus.error is None

        with Ignore(ZeroDivisionError, if_else=run_function) as relay:
            pass
        assert relay.bus.error is None

    async def test_if_else_doesnt_run_when_exceptions(self) -> None:
        async with Ignore(
            ZeroDivisionError, if_else=arun_function
        ) as relay:
            raise ZeroDivisionError
        assert not hasattr(relay.bus, "error")

        with Ignore(ZeroDivisionError, if_else=run_function) as relay:
            raise ZeroDivisionError
        assert not hasattr(relay.bus, "error")

    async def test_if_except_doesnt_run_when_no_exceptions(self) -> None:
        async with Ignore(
            ZeroDivisionError, if_except=arun_function
        ) as relay:
            pass
        assert not hasattr(relay.bus, "error")

        with Ignore(ZeroDivisionError, if_except=run_function) as relay:
            pass
        assert not hasattr(relay.bus, "error")

    async def test_if_except_doesnt_run_when_wrong_exceptions(self) -> None:
        try:
            async with Ignore(
                ZeroDivisionError, if_except=arun_function
            ) as relay:
                raise TypeError
        except TypeError:
            relay.bus.wrong_exception_not_handled = True
        assert not hasattr(relay.bus, "error")
        assert getattr(relay.bus, "wrong_exception_not_handled", False)

        try:
            with Ignore(ZeroDivisionError, if_except=run_function) as relay:
                raise TypeError
        except TypeError:
            relay.bus.wrong_exception_not_handled = True
        assert not hasattr(relay.bus, "error")
        assert getattr(relay.bus, "wrong_exception_not_handled", False)

    async def test_if_except_runs_when_correct_exceptions(self) -> None:
        async with Ignore(
            ZeroDivisionError, if_except=arun_function
        ) as relay:
            raise ZeroDivisionError
        assert relay.bus.error.__class__ is ZeroDivisionError

        with Ignore(ZeroDivisionError, if_except=run_function) as relay:
            raise ZeroDivisionError
        assert relay.bus.error.__class__ is ZeroDivisionError

    async def test_if_except_returns_control_bool_when_correct_exceptions(
        self,
    ) -> None:
        try:
            async with Ignore(
                ZeroDivisionError, if_except=arun_function_return_false
            ) as relay:
                raise ZeroDivisionError
        except ZeroDivisionError as exception:
            error = exception
        assert relay.bus.error is error

        try:
            with Ignore(
                ZeroDivisionError, if_except=run_function_return_false
            ) as relay:
                raise ZeroDivisionError
        except ZeroDivisionError as exception:
            error = exception
        assert relay.bus.error is error

    async def test_finally_run_goes_after_if_else(self) -> None:
        async with Ignore(
            ZeroDivisionError,
            if_else=arun_function_order_checking(name="if_else"),
            if_except=arun_function_order_checking(name="if_except"),
            finally_run=arun_function_order_checking(name="finally_run"),
        ) as relay:
            relay.bus.count = 0
        assert relay.bus.count == 2
        assert relay.bus.if_else == 0
        assert not hasattr(relay.bus, "if_except")
        assert relay.bus.finally_run == 1

        with Ignore(
            ZeroDivisionError,
            if_else=run_function_order_checking(name="if_else"),
            if_except=arun_function_order_checking(name="if_except"),
            finally_run=run_function_order_checking(name="finally_run"),
        ) as relay:
            relay.bus.count = 0
        assert relay.bus.count == 2
        assert relay.bus.if_else == 0
        assert not hasattr(relay.bus, "if_except")
        assert relay.bus.finally_run == 1

    async def test_finally_run_goes_after_if_except(self) -> None:
        async with Ignore(
            ZeroDivisionError,
            if_else=run_function_order_checking(name="if_else"),
            if_except=arun_function_order_checking(name="if_except"),
            finally_run=arun_function_order_checking(name="finally_run"),
        ) as relay:
            relay.bus.count = 0
            raise ZeroDivisionError
        assert relay.bus.count == 2
        assert not hasattr(relay.bus, "if_else")
        assert relay.bus.if_except == 0
        assert relay.bus.finally_run == 1

        with Ignore(
            ZeroDivisionError,
            if_else=run_function_order_checking(name="if_else"),
            if_except=run_function_order_checking(name="if_except"),
            finally_run=run_function_order_checking(name="finally_run"),
        ) as relay:
            relay.bus.count = 0
            raise ZeroDivisionError
        assert relay.bus.count == 2
        assert not hasattr(relay.bus, "if_else")
        assert relay.bus.if_except == 0
        assert relay.bus.finally_run == 1

    async def test_raised_error_is_shown_in_object_repr(self) -> None:
        with Ignore(ZeroDivisionError) as ignored:
            relay = ignored
            raise ZeroDivisionError
        assert "ZeroDivisionError" in repr(relay)


__all__ = sorted({n for n in globals() if n.lower().startswith("test")})



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


import multiprocessing


try:
    multiprocessing.set_start_method("fork", force=True)
except ValueError:
    multiprocessing.set_start_method("spawn", force=True)
multiprocessing.freeze_support()


import sys
import json
import pytest
import hashlib
import builtins
from pathlib import Path
from functools import partial
from hashlib import sha3_256, sha3_512, shake_128, shake_256


_PACKAGE_PATH = str(Path(__file__).absolute().parent.parent)
sys.path.insert(0, _PACKAGE_PATH)


import aiootp
from aiootp import *
from aiootp._paths import *
from aiootp._typing import *
from aiootp._constants.misc import *
from aiootp._constants.datasets import *
from aiootp._exceptions import *
from aiootp._gentools import aunpack, unpack, batch, abatch
from aiootp._permutations import *
from aiootp._debug_control import DebugControl
from aiootp.commons.slots import Slots, FrozenSlots
from aiootp.commons.namespaces import Namespace, OpenNamespace
from aiootp.commons.configs import Config, ConfigMap
from aiootp.asynchs import Processes, Threads, Clock
from aiootp.asynchs import run, asleep, get_process_id, get_thread_id
from aiootp.generics import *
from aiootp.generics import ByteIO
from aiootp.generics.canon import acanonical_pack, canonical_pack
from aiootp.generics.canon import acanonical_unpack, canonical_unpack
from aiootp.generics.canon import aencode_key, encode_key
from aiootp.generics.transform import axi_mix, xi_mix
from aiootp.randoms import *
from aiootp.randoms.simple import *
from aiootp.ciphers import Ciphertext, Padding


# Use enable_debugging method to run tests in debug mode. May cause noticable slowdown.
DebugControl.disable_debugging()


# create static values for this test session
t = Typing

violation = lambda problem: lambda relay: raise_exception(
    AssertionError(f"{problem} : {relay!r}")
)


class MemoizedCipher(FrozenSlots):
    __slots__ = ("config", "cipher", "salt", "aad")

    def __init__(self, cipher_type: type) -> None:
        self.config = cipher_type._config
        self.cipher = cipher_type(key=key)
        self.salt = csprng(self.config.SALT_BYTES)
        self.aad = f"testing {self.config.NAME}".encode()

    def __iter__(
        self,
    ) -> t.Generator[
        None, t.Union[t.ConfigType, t.CipherInterfaceType, bytes], None
    ]:
        yield self.config
        yield self.cipher
        yield self.salt
        yield self.aad


key = csprng(168)
dual_output_ciphers = [MemoizedCipher(Chunky2048)]
shake_permute_ciphers = [MemoizedCipher(Slick256)]
all_ciphers = [*dual_output_ciphers, *shake_permute_ciphers]
dual_output_cipher_names = [conf.NAME for (conf, *_) in dual_output_ciphers]
shake_permute_cipher_names = [
    conf.NAME for (conf, *_) in shake_permute_ciphers
]


chunky2048_test_vector_0 = OpenNamespace(
    plaintext=(
        b"it is not a doctrine to be preached, but a deed to be done."
    ),
    key=bytes.fromhex(
        "54868dc506d99611adb67f1b41eda44fd7151e135860a6d791cad6df6715bd"
        "5204713bc2b064e88a429b93aaa282d9cc7c0c5cdaaffb65cc268b7acedd5e"
        "531a"
    ),
    shmac=bytes.fromhex(
        "f791571464dd45754e16d8e6dea2f1a05204dca2510589be012bc5117dfd98"
        "e8"
    ),
    salt=bytes.fromhex("cf3761d57860bdbc"),
    iv=bytes.fromhex("e09d2eee03ae82e6"),
    aad=b"test_vector_I",
    ciphertext=bytes.fromhex(
        "1ce9033362f402390f872fa4f20de14d8a13f32f3906dd302a275b2c62f9f5"
        "f1a8ddb61871a2bab61dffaa2b414c2ca68601a61d9670a944619fc5105469"
        "b95e63c494326678e865a751d2ab0cac40439aa51cec3a8aa9e4cf0255ea96"
        "659a4cce41e2c8dac09f97d5bb05efb39bcf6043c6cd18364f0ecc0c721160"
        "e8204e333b1e336c55bd9a09ba1c0ce08805cc924c1925ad238b785ad17e6c"
        "064b2fe8676dc2fff3a3a895381cbf8d5aa04970c4f210ae323a5274c8f703"
        "4167ba0f7042e17d748d68327d2219c83bf4e4fb9f5d36c0b0c5ca661db09c"
        "b887d2bd39589d2e12f7a799b2e3e1cad7a7dbfd4c9d1a46de6ed5e51d7561"
        "70de526f4c4fa9fb"
    ),
)


slick256_test_vector_0 = OpenNamespace(
    plaintext=(
        b"it is not a doctrine to be preached, but a deed to be done."
    ),
    key=bytes.fromhex(
        "54868dc506d99611adb67f1b41eda44fd7151e135860a6d791cad6df6715bd"
        "5204713bc2b064e88a429b93aaa282d9cc7c0c5cdaaffb65cc268b7acedd5e"
        "531a"
    ),
    shmac=bytes.fromhex("5e6011b3b1ae4314b4192a1e6a18b2ca8130bc5cafcd7287"),
    salt=bytes.fromhex("cf3761d57860bdbc"),
    iv=bytes.fromhex("dda31acd46832df9"),
    aad=b"test_vector_I",
    ciphertext=bytes.fromhex(
        "f73e7361a73af3eeabc88247555e06fc5bdaea10482c8351eab3ee9cfcc6fe"
        "f678b1171e6458fcef3d27896bee262769a742f8ffce51cd44db2a8e5673f4"
        "5770df4b907683c881049f15a92bed514613378ec4beb07766e809b5f68e1e"
        "326bc7"
    ),
)


passphrase_0 = csprng()
passphrase_1 = randoms.token_bytes(32)
passphrases = [passphrase_0, passphrase_1]

salt_0 = csprng()
salt_1 = randoms.token_bytes(32)
salts = [salt_0, salt_1]

passcrypt_settings = OpenNamespace(mb=1, cpu=1, cores=1)

tag = "testing"
atag = "a" + tag
metatag = "clients"
ametatag = "a" + metatag

username = b"test suite"
passphrase = b"terrible low entropy passphrase"
PROFILE = dict(
    username=username,
    passphrase=passphrase,
    salt=salt_0,
    aad=b"testing passcrypt",
)
LOW_PASSCRYPT_SETTINGS = OpenNamespace(mb=1, cpu=1, cores=1)
PROFILE_AND_SETTINGS = {**PROFILE, **LOW_PASSCRYPT_SETTINGS}

passcrypt_test_vector_0 = OpenNamespace(
    mb=2,
    cpu=2,
    cores=2,
    tag_size=257,
    salt_size=256,
    aad=b"",
    timestamp=bytes.fromhex("000a59139c603129"),
    passphrase=b"test vector passphrase I",
    salt=bytes.fromhex(
        "b3c100b7670ef0c9e15bf877a80d78342ab1f6039511d8444ef82ee9f185fd"
        "ff3bebf168f4af41c5d735a2a6a19c8aab2d142353e307968d9d65e33c9424"
        "a51e0b70c4415723d5c72267cd992fd51141c2ce6ae94ccff9945d1b222fc0"
        "0be81504032ab3e51d0ed617d51fee6cfd5c5302bbd7ada60e0f3d45d631f6"
        "a7c4dc47517f20659a1c00f0f371d8a2e09bb247d4493502b4460f3826a094"
        "ac50faa39b8473629ae782d7fb3bb6c69dfaf68a52f06bc3f92d5c88d99a1f"
        "064d377dd894ee14993cf4d72cd78a4401d519a78fca5b2d7b4b6557e75507"
        "ab94021f2fcb29a10d87c0deca93607e339c1dd67d2db939ee926028712073"
        "bff2135410a51bb2"
    ),
    tag=bytes.fromhex(
        "4b3b9a627021735a7a0c2f595bab62135b2fb918711309da9e98e5336fa18a"
        "394e165a9d2a00750d4fd52d4e26562f44d1850327e7e583f1f69b35ea4483"
        "3f00efc01cdfebfeca15f09e377a786ce4a3efa8b0e945f4b07a420bc2c878"
        "c6f29bd2e0dd6156a9575450c913722606cf03bb2f9a951e384e17638a3b16"
        "0cb88199a7c3f099e7be390101e1933662c523e1cd61ba78d77a15607a1b3f"
        "7ec2f53486c8537fccba1da9fff7aac32c65170237b429de52a9f596cc8ea9"
        "95a877503ffd6e04caaef3ceacc1cdc0828243e37d82d263517f5336567a19"
        "54e5515a7a33a820f620b118b74c7089ec1f7c00207659b9b5a737db7cb1ca"
        "aaa01a51b2b8641014"
    ),
    hash_passphrase_result=bytes.fromhex(
        "000a59139c6031290000010101ffb3c100b7670ef0c9e15bf877a80d78342a"
        "b1f6039511d8444ef82ee9f185fdff3bebf168f4af41c5d735a2a6a19c8aab"
        "2d142353e307968d9d65e33c9424a51e0b70c4415723d5c72267cd992fd511"
        "41c2ce6ae94ccff9945d1b222fc00be81504032ab3e51d0ed617d51fee6cfd"
        "5c5302bbd7ada60e0f3d45d631f6a7c4dc47517f20659a1c00f0f371d8a2e0"
        "9bb247d4493502b4460f3826a094ac50faa39b8473629ae782d7fb3bb6c69d"
        "faf68a52f06bc3f92d5c88d99a1f064d377dd894ee14993cf4d72cd78a4401"
        "d519a78fca5b2d7b4b6557e75507ab94021f2fcb29a10d87c0deca93607e33"
        "9c1dd67d2db939ee926028712073bff2135410a51bb24b3b9a627021735a7a"
        "0c2f595bab62135b2fb918711309da9e98e5336fa18a394e165a9d2a00750d"
        "4fd52d4e26562f44d1850327e7e583f1f69b35ea44833f00efc01cdfebfeca"
        "15f09e377a786ce4a3efa8b0e945f4b07a420bc2c878c6f29bd2e0dd6156a9"
        "575450c913722606cf03bb2f9a951e384e17638a3b160cb88199a7c3f099e7"
        "be390101e1933662c523e1cd61ba78d77a15607a1b3f7ec2f53486c8537fcc"
        "ba1da9fff7aac32c65170237b429de52a9f596cc8ea995a877503ffd6e04ca"
        "aef3ceacc1cdc0828243e37d82d263517f5336567a1954e5515a7a33a820f6"
        "20b118b74c7089ec1f7c00207659b9b5a737db7cb1caaaa01a51b2b8641014"
    ),
)

passcrypt_test_vector_1 = OpenNamespace(
    mb=3,
    cpu=3,
    cores=3,
    tag_size=32,
    salt_size=10,
    aad=b"testvector",
    timestamp=bytes.fromhex("000a59328da314c9"),
    passphrase=b"test vector passphrase II",
    salt=bytes.fromhex("ca926bc906fa14b886eb"),
    tag=bytes.fromhex(
        "597dd54fabbeddd3605b3c2bd4ac2c476ac457978ddabf31ccd4a3b8f33942"
        "e8"
    ),
    hash_passphrase_result=bytes.fromhex(
        "000a59328da314c9000002020209ca926bc906fa14b886eb597dd54fabbedd"
        "d3605b3c2bd4ac2c476ac457978ddabf31ccd4a3b8f33942e8"
    ),
)

passcrypt_test_vector_2 = OpenNamespace(
    mb=4,
    cpu=4,
    cores=4,
    tag_size=16,
    salt_size=4,
    aad=b"core_cache_change",
    timestamp=bytes.fromhex("000a59529a517df8"),
    passphrase=b"test vector passphrase III",
    salt=bytes.fromhex("80345361"),
    tag=bytes.fromhex("4a9f3a8ad7b60afcdc2400fd3e0f3ff4"),
    hash_passphrase_result=bytes.fromhex(
        "000a59529a517df8000003030303803453614a9f3a8ad7b60afcdc2400fd3e"
        "0f3ff4"
    ),
)

passcrypt_test_vector_3 = OpenNamespace(
    mb=1,
    cpu=5,
    cores=5,
    tag_size=24,
    salt_size=8,
    aad=bytes.fromhex(
        "000102030405060708090a0b0c0d0e0f101112131415161718191a1b1c1d1e"
        "1f202122232425262728292a2b2c2d2e2f303132333435363738393a3b3c3d"
        "3e3f404142434445464748494a4b4c4d4e4f505152535455565758595a5b5c"
        "5d5e5f606162636465666768696a6b6c6d6e6f707172737475767778797a7b"
        "7c7d7e7f808182838485868788898a8b8c8d8e8f909192939495969798999a"
        "9b9c9d9e9fa0a1a2a3a4a5a6a7a8a9aaabacadaeafb0b1b2b3b4b5b6b7b8b9"
        "babbbcbdbebfc0c1c2c3c4c5c6c7c8c9cacbcccdcecfd0d1d2d3d4d5d6d7d8"
        "d9dadbdcdddedfe0e1e2e3e4e5e6e7e8e9eaebecedeeeff0f1f2f3f4f5f6f7"
        "f8f9fafbfcfdfeff"
    ),
    timestamp=bytes.fromhex("000a59a53cfe5ecd"),
    passphrase=b"test vector passphrase IIII",
    salt=bytes.fromhex("fbd2ee954afb48c3"),
    tag=bytes.fromhex("527a0138c531ea8c027fcad351fd11674c686a16265b5368"),
    hash_passphrase_result=bytes.fromhex(
        "000a59a53cfe5ecd000000040407fbd2ee954afb48c3527a0138c531ea8c02"
        "7fcad351fd11674c686a16265b5368"
    ),
)

byte_leakage = 16 * b"\x00"
string_leakage = 16 * b"0"
plaintext_bytes = b"\xff" + 256 * b"\x00"
plaintext_string = "f" + 256 * "0"
test_data = {
    "floats": 10000.243,
    "ints": list(range(-16, 16, 8)),
    "dicts": {"testing": {}},
    "lists": 4 * [[]],
    "strings": "testing...",
    "nul": None,
    "bools": [True, False],
}
atest_data = test_data.copy()


# Creating timestamped values early so later testing has to wait less time
ttl_test_cipher = choice(all_ciphers)
atest_json_ciphertext = run(
    ttl_test_cipher.cipher.ajson_encrypt(
        atest_data, aad=ttl_test_cipher.aad
    )
)
test_json_ciphertext = ttl_test_cipher.cipher.json_encrypt(
    test_data, aad=ttl_test_cipher.aad
)
atest_token_ciphertext = run(
    ttl_test_cipher.cipher.amake_token(
        plaintext_bytes, aad=ttl_test_cipher.aad
    )
)
test_token_ciphertext = ttl_test_cipher.cipher.make_token(
    plaintext_bytes, aad=ttl_test_cipher.aad
)


async def make_async_ttl_cipher_stream() -> (
    t.Tuple[t.AsyncCipherStream, bytes]
):
    stream = await ttl_test_cipher.cipher.astream_encrypt(
        salt=ttl_test_cipher.salt, aad=ttl_test_cipher.aad
    )
    await stream.abuffer(plaintext_bytes)
    ciphertext = b"".join(
        [(block_id + block) async for block_id, block in stream]
    )
    ciphertext += b"".join(
        [(block_id + block) async for block_id, block in stream.afinalize()]
    )
    return stream, ciphertext


attl_cipher_stream, attl_stream_ciphertext = run(
    make_async_ttl_cipher_stream()
)

ttl_cipher_stream = ttl_test_cipher.cipher.stream_encrypt(
    salt=ttl_test_cipher.salt, aad=ttl_test_cipher.aad
)
ttl_cipher_stream.buffer(plaintext_bytes)
ttl_stream_ciphertext = b"".join(
    (block_id + block) for block_id, block in ttl_cipher_stream.finalize()
)

light_pcrypt = Passcrypt(mb=1, cpu=1, cores=1, tag_size=32)
aexpired_passcrypt_hash = run(light_pcrypt.ahash_passphrase(passphrase_0))
expired_passcrypt_hash = light_pcrypt.hash_passphrase(passphrase_0)

clock = Clock(SECONDS, epoch=EPOCH_NS)
ns_clock = Clock(NANOSECONDS, epoch=EPOCH_NS)
time_start = clock.make_timestamp()


@pytest.fixture(scope="session")
def database():
    db = Database(key=key, preload=True)
    db.save_database()
    yield db

    db.delete_database()


@pytest.fixture(scope="session")
def async_database():
    db = run(AsyncDatabase(key=key, preload=True))
    yield db

    run(db.adelete_database())


@pytest.fixture()
def path():
    file_path = Path("byte_io_testing_path.txt").absolute()

    if not file_path.is_file():
        file_path.write_bytes(b"")

    yield file_path

    if file_path.is_file():
        file_path.unlink()


class ExampleConfig(Config):
    __slots__ = ("NUMBER", "STRING")

    slots_types = dict(NUMBER=int, STRING=str)

    def __init__(self, number: int, string: str) -> None:
        self.NUMBER = number
        self.STRING = string


@pytest.fixture()
def config():
    return ExampleConfig(number=420, string="word")


@pytest.fixture()
def mapping():
    return ConfigMap(config_type=ExampleConfig)


__all__ = [n for n in globals() if not n.startswith("_")]



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


from conftest import *


class TestTypingClass:
    async def test_added_type_names_must_have_valid_identifiers(
        self,
    ) -> None:
        problem = (  # fmt: skip
            "A type with non-identifier name was allowed to be registered."
        )
        NonIdentifierNamedType = type("123Type", (), {})
        with Ignore(ValueError, if_else=violation(problem)):
            t.add_type(NonIdentifierNamedType)

    async def test_cant_register_type_more_than_once(self) -> None:
        problem = (  # fmt: skip
            "A type was registered more than once."
        )
        with Ignore(AttributeError, if_else=violation(problem)):
            t.add_type(Domains)

    async def test_mixed_case_type_cant_be_registered(self) -> None:
        problem = (  # fmt: skip
            "A mixed-case type was registered."
        )
        Mixed_Case_Type = type("Mixed_Case_Type", (), {})
        with Ignore(ValueError, if_else=violation(problem)):
            t.add_type(Mixed_Case_Type)

    async def test_non_capitalized_type_cant_be_registered(self) -> None:
        problem = (  # fmt: skip
            "A non-capitalized type was registered."
        )
        nonCapitalizedType = type("nonCapitalizedType", (), {})
        with Ignore(ValueError, if_else=violation(problem)):
            t.add_type(nonCapitalizedType)

    async def test_non_class_object_cant_be_registered(self) -> None:
        problem = (  # fmt: skip
            "A non-class object was registered."
        )
        non_class_objects = (
            1,
            1.1,
            None,
            "test",
            b"test",
            [],
            {},
            t.PositiveRealNumber,
        )
        errors = (AttributeError, ValueError, TypeError)
        for obj in non_class_objects:
            with Ignore(*errors, if_else=violation(problem)):
                t.add_type(obj)


__all__ = sorted({n for n in globals() if n.lower().startswith("test")})



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


from conftest import *

from aiootp._constants.misc import B_TO_MB_RATIO
from aiootp.keygens.passcrypt import PasscryptHash
from aiootp.keygens.passcrypt.sessions_manager import PasscryptProcesses
from aiootp.keygens.passcrypt.config import passcrypt_spec as config


PasscryptSession = t.PasscryptSession


class TestPasscryptMetadataHashes:
    passphrase = b"a generic passphrase 123456"
    pcrypt = Passcrypt(**passcrypt_settings, tag_size=16)
    config = pcrypt._config
    ametadata_hash = run(pcrypt.ahash_passphrase(passphrase))
    metadata_hash = pcrypt.hash_passphrase(passphrase)

    async def test_hashes_are_unique(self) -> None:
        tag_slice = slice(-self.pcrypt._settings.tag_size, None)
        assert self.metadata_hash != self.ametadata_hash
        assert (
            self.metadata_hash[tag_slice] != self.ametadata_hash[tag_slice]
        )

    async def test_hashes_are_bytes(self) -> None:
        assert type(self.ametadata_hash) is bytes
        assert type(self.metadata_hash) is bytes

    async def test_async_hash_reconstruction(self) -> None:
        apcrypt_hash = PasscryptHash(config=self.config).import_hash(
            self.ametadata_hash
        )
        ahash_check = await self.pcrypt.anew(
            self.passphrase, apcrypt_hash.salt, aad=apcrypt_hash.timestamp
        )
        assert apcrypt_hash.tag_size == len(ahash_check)
        assert ahash_check == self.ametadata_hash[-apcrypt_hash.tag_size :]
        assert ahash_check == self.pcrypt.new(
            self.passphrase, apcrypt_hash.salt, aad=apcrypt_hash.timestamp
        )
        assert {**apcrypt_hash} == dict(
            timestamp=apcrypt_hash.timestamp,
            mb=apcrypt_hash.mb,
            cpu=apcrypt_hash.cpu,
            cores=apcrypt_hash.cores,
            salt=apcrypt_hash.salt,
            tag=apcrypt_hash.tag,
        )
        assert self.ametadata_hash == (
            apcrypt_hash.timestamp
            + (apcrypt_hash.mb - 1).to_bytes(config.MB_BYTES, BIG)
            + (apcrypt_hash.cpu - 1).to_bytes(config.CPU_BYTES, BIG)
            + (apcrypt_hash.cores - 1).to_bytes(config.CORES_BYTES, BIG)
            + (len(apcrypt_hash.salt) - 1).to_bytes(
                config.SALT_SIZE_BYTES, BIG
            )
            + apcrypt_hash.salt
            + apcrypt_hash.tag
        )

    async def test_sync_hash_reconstruction(self) -> None:
        pcrypt_hash = PasscryptHash(config=self.config).import_hash(
            self.metadata_hash
        )
        hash_check = self.pcrypt.new(
            self.passphrase, pcrypt_hash.salt, aad=pcrypt_hash.timestamp
        )
        assert pcrypt_hash.tag_size == len(hash_check)
        assert hash_check == self.metadata_hash[-pcrypt_hash.tag_size :]
        assert hash_check == await self.pcrypt.anew(
            self.passphrase, pcrypt_hash.salt, aad=pcrypt_hash.timestamp
        )
        assert {**pcrypt_hash} == dict(
            timestamp=pcrypt_hash.timestamp,
            mb=pcrypt_hash.mb,
            cpu=pcrypt_hash.cpu,
            cores=pcrypt_hash.cores,
            salt=pcrypt_hash.salt,
            tag=pcrypt_hash.tag,
        )
        # fmt: off
        assert self.metadata_hash == (
            pcrypt_hash.timestamp
            + (pcrypt_hash.mb - 1).to_bytes(config.MB_BYTES, BIG)
            + (pcrypt_hash.cpu - 1).to_bytes(config.CPU_BYTES, BIG)
            + (pcrypt_hash.cores - 1).to_bytes(config.CORES_BYTES, BIG)
            + (len(pcrypt_hash.salt) - 1).to_bytes(config.SALT_SIZE_BYTES, BIG)
            + pcrypt_hash.salt
            + pcrypt_hash.tag
        )
        # fmt: on

    async def test_verify_methods_detect_correct_passphrase(self) -> None:
        await Passcrypt.averify(
            self.ametadata_hash, self.passphrase, config=self.config
        )
        Passcrypt.verify(
            self.metadata_hash, self.passphrase, config=self.config
        )

    async def test_wrong_passphrase_fails_async(self) -> None:
        problem = (  # fmt: skip
            "An invalid passphrase passed async verification."
        )
        with Ignore(
            Passcrypt.InvalidPassphrase, if_else=violation(problem)
        ):
            await Passcrypt.averify(
                self.ametadata_hash,
                self.passphrase + b"\x00",
                config=self.config,
            )

    async def test_wrong_passphrase_fails_sync(self) -> None:
        problem = (  # fmt: skip
            "An invalid passphrase passed sync verification."
        )
        with Ignore(
            Passcrypt.InvalidPassphrase, if_else=violation(problem)
        ):
            Passcrypt.verify(
                self.metadata_hash,
                self.passphrase + b"\x00",
                config=self.config,
            )

    async def test_async_mb_resource_limitations(self) -> None:
        mb_allowed = passcrypt_settings.mb
        problem = (  # fmt: skip
            "The `mb` was allowed to exceed the resource limit of "
            f"{mb_allowed - 1}."
        )
        with Ignore(ResourceWarning, if_else=violation(problem)):
            await Passcrypt.averify(
                self.ametadata_hash,
                self.passphrase,
                mb_allowed=range(1, mb_allowed - 1),
                config=self.config,
            )

        problem = (  # fmt: skip
            "The `mb` was allowed to fall below the resource limit of "
            f"{mb_allowed + 1}."
        )
        with Ignore(ResourceWarning, if_else=violation(problem)):
            await Passcrypt.averify(
                self.ametadata_hash,
                self.passphrase,
                mb_allowed=range(mb_allowed + 1, mb_allowed + 2),
                config=self.config,
            )

    async def test_sync_mb_resource_limitations(self) -> None:
        mb_allowed = passcrypt_settings.mb
        problem = (  # fmt: skip
            "The `mb` was allowed to exceed the resource limit of "
            f"{mb_allowed - 1}."
        )
        with Ignore(ResourceWarning, if_else=violation(problem)):
            Passcrypt.verify(
                self.metadata_hash,
                self.passphrase,
                mb_allowed=range(1, mb_allowed - 1),
                config=self.config,
            )

        problem = (  # fmt: skip
            "The `mb` was allowed to fall below the resource limit of "
            f"{mb_allowed + 1}."
        )
        with Ignore(ResourceWarning, if_else=violation(problem)):
            Passcrypt.verify(
                self.metadata_hash,
                self.passphrase,
                mb_allowed=range(mb_allowed + 1, mb_allowed + 2),
                config=self.config,
            )

    async def test_async_cpu_resource_limitations(self) -> None:
        cpu_allowed = passcrypt_settings.cpu
        problem = (  # fmt: skip
            "The `cpu` was allowed to exceed the resource limit of "
            f"{cpu_allowed - 1}."
        )
        with Ignore(ResourceWarning, if_else=violation(problem)):
            await Passcrypt.averify(
                self.ametadata_hash,
                self.passphrase,
                cpu_allowed=range(256, cpu_allowed - 1),
                config=self.config,
            )

        problem = (  # fmt: skip
            "The `cpu` was allowed to fall below the resource limit of "
            f"{cpu_allowed + 1}."
        )
        with Ignore(ResourceWarning, if_else=violation(problem)):
            await Passcrypt.averify(
                self.ametadata_hash,
                self.passphrase,
                cpu_allowed=range(cpu_allowed + 1, 257),
                config=self.config,
            )

    async def test_sync_cpu_resource_limitations(self) -> None:
        cpu_allowed = passcrypt_settings.cpu
        problem = (  # fmt: skip
            "The `cpu` was allowed to exceed the resource limit of "
            f"{cpu_allowed - 1}."
        )
        with Ignore(ResourceWarning, if_else=violation(problem)):
            Passcrypt.verify(
                self.metadata_hash,
                self.passphrase,
                cpu_allowed=range(256, cpu_allowed - 1),
                config=self.config,
            )

        problem = (  # fmt: skip
            "The `cpu` was allowed to fall below the resource limit of "
            f"{cpu_allowed + 1}."
        )
        with Ignore(ResourceWarning, if_else=violation(problem)):
            Passcrypt.verify(
                self.metadata_hash,
                self.passphrase,
                cpu_allowed=range(cpu_allowed + 1, 257),
                config=self.config,
            )

    async def test_async_cores_resource_limitations(self) -> None:
        cores_allowed = passcrypt_settings.cores
        problem = (  # fmt: skip
            "The `cores` was allowed to exceed the resource limit of "
            f"{cores_allowed - 1}."
        )
        with Ignore(ResourceWarning, if_else=violation(problem)):
            await Passcrypt.averify(
                self.ametadata_hash,
                self.passphrase,
                cores_allowed=range(256, cores_allowed - 1),
                config=self.config,
            )

        problem = (  # fmt: skip
            "The `cores` was allowed to fall below the resource limit of "
            f"{cores_allowed + 1}."
        )
        with Ignore(ResourceWarning, if_else=violation(problem)):
            await Passcrypt.averify(
                self.ametadata_hash,
                self.passphrase,
                cores_allowed=range(cores_allowed + 1, 257),
                config=self.config,
            )

    async def test_sync_cores_resource_limitations(self) -> None:
        cores_allowed = passcrypt_settings.cores
        problem = (  # fmt: skip
            "The `cores` was allowed to exceed the resource limit of "
            f"{cores_allowed - 1}."
        )
        with Ignore(ResourceWarning, if_else=violation(problem)):
            Passcrypt.verify(
                self.metadata_hash,
                self.passphrase,
                cores_allowed=range(256, cores_allowed - 1),
                config=self.config,
            )

        problem = (  # fmt: skip
            "The `cores` was allowed to fall below the resource limit of "
            f"{cores_allowed + 1}."
        )
        with Ignore(ResourceWarning, if_else=violation(problem)):
            Passcrypt.verify(
                self.metadata_hash,
                self.passphrase,
                cores_allowed=range(cores_allowed + 1, 257),
                config=self.config,
            )


class TestPasscryptTestVectors:
    def accrue_failures(
        self, index: int, failures: t.List[Exception]
    ) -> None:
        return lambda relay: failures.append((index, relay.error)) or True

    async def test_declared_algorithm_example_results(self) -> None:
        example_results = (
            passcrypt_test_vector_0,
            passcrypt_test_vector_1,
            passcrypt_test_vector_2,
            passcrypt_test_vector_3,
        )
        failures = []
        for index, result in enumerate(example_results):
            with Ignore(
                AssertionError,
                if_except=self.accrue_failures(index, failures),
            ):
                await self.examples_match_derivation(result)
        assert not any(failures)

    async def examples_match_derivation(self, result) -> None:
        assert result.hash_passphrase_result == (
            result.timestamp
            + (result.mb - 1).to_bytes(config.MB_BYTES, BIG)
            + (result.cpu - 1).to_bytes(config.CPU_BYTES, BIG)
            + (result.cores - 1).to_bytes(config.CORES_BYTES, BIG)
            + (result.salt_size - 1).to_bytes(config.SALT_SIZE_BYTES, BIG)
            + result.salt
            + result.tag
        )
        pcrypt = Passcrypt(
            mb=result.mb,
            cpu=result.cpu,
            cores=result.cores,
            tag_size=result.tag_size,
        )
        assert result.tag == pcrypt.new(
            result.passphrase,
            result.salt,
            aad=result.timestamp + result.aad,
        )
        await Passcrypt.averify(
            result.hash_passphrase_result,
            result.passphrase,
            aad=result.aad,
            config=pcrypt._config,
        )


class TestPasscryptInputsOutputs:
    pcrypt = Passcrypt(
        mb=config.MIN_MB,
        cpu=config.MIN_CPU,
        cores=config.MIN_CORES,
        tag_size=config.MIN_TAG_SIZE,
        salt_size=config.MIN_SALT_SIZE,
    )
    salt = csprng(config.MIN_SALT_SIZE)
    object.__delattr__(pcrypt._settings, "salt_size")
    config = pcrypt._config
    settings = dict(**pcrypt._settings, config=pcrypt._config)

    async def test_empty_passphrase_isnt_allowed(self) -> None:
        problem = (  # fmt: skip
            "Empty passphrase was allowed."
        )
        with Ignore(
            Passcrypt.ImproperPassphrase, if_else=violation(problem)
        ):
            PasscryptSession(b"", self.salt, aad=b"", **self.settings)

        with Ignore(
            Passcrypt.ImproperPassphrase, if_else=violation(problem)
        ):
            self.pcrypt.new(b"", self.salt)

        async with Ignore(
            Passcrypt.ImproperPassphrase, if_else=violation(problem)
        ):
            await self.pcrypt.anew(b"", self.salt)

    async def test_min_passphrase_length(self) -> None:
        problem = (  # fmt: skip
            "A below minimum length passphrase was allowed."
        )
        with Ignore(ValueError, if_else=violation(problem)):
            PasscryptSession(
                (config.MIN_PASSPHRASE_BYTES - 1) * b"p",
                self.salt,
                **self.settings,
            )

        with Ignore(ValueError, if_else=violation(problem)):
            self.pcrypt.new(
                (config.MIN_PASSPHRASE_BYTES - 1) * b"p", self.salt
            )

        async with Ignore(ValueError, if_else=violation(problem)):
            await self.pcrypt.anew(
                (config.MIN_PASSPHRASE_BYTES - 1) * b"p", self.salt
            )

    async def test_passphrase_must_be_bytes(self) -> None:
        problem = (  # fmt: skip
            "Non-bytes passphrase was allowed."
        )
        with Ignore(TypeError, if_else=violation(problem)):
            PasscryptSession(
                "a string passphrase", self.salt, **self.settings
            )

        with Ignore(TypeError, if_else=violation(problem)):
            self.pcrypt.new("a string passphrase", self.salt)

        with Ignore(TypeError, if_else=violation(problem)):
            self.pcrypt.new(None, self.salt)

        with Ignore(TypeError, if_else=violation(problem)):
            self.pcrypt.new(12345, self.salt)

        async with Ignore(TypeError, if_else=violation(problem)):
            await self.pcrypt.anew("a string passphrase", self.salt)

    async def test_falsey_salts_not_allowed(self) -> None:
        problem = (  # fmt: skip
            "Empty salt was allowed."
        )
        with Ignore(ValueError, if_else=violation(problem)):
            PasscryptSession(key, salt=b"", **self.settings)

        with Ignore(ValueError, if_else=violation(problem)):
            self.pcrypt.new(key, salt=b"")

        async with Ignore(ValueError, if_else=violation(problem)):
            await self.pcrypt.anew(key, salt=b"")

    async def test_salt_must_be_bytes(self) -> None:
        problem = (  # fmt: skip
            "Non-bytes salt was allowed."
        )
        with Ignore(TypeError, if_else=violation(problem)):
            PasscryptSession(key, "a string salt here", **self.settings)

        with Ignore(TypeError, if_else=violation(problem)):
            self.pcrypt.new(key, "a string salt here")

        with Ignore(TypeError, if_else=violation(problem)):
            self.pcrypt.new(key, None)

        with Ignore(TypeError, if_else=violation(problem)):
            self.pcrypt.new(key, 123456)

        async with Ignore(TypeError, if_else=violation(problem)):
            await self.pcrypt.anew(key, "a string salt here")

    async def test_aad_must_be_bytes(self) -> None:
        problem = (  # fmt: skip
            "A non-bytes type `aad` was allowed."
        )
        with Ignore(TypeError, if_else=violation(problem)):
            PasscryptSession(
                key, self.salt, aad="some string aad", **self.settings
            )

        with Ignore(TypeError, if_else=violation(problem)):
            self.pcrypt.new(key, self.salt, aad=None)

        with Ignore(TypeError, if_else=violation(problem)):
            self.pcrypt.new(key, self.salt, aad=12345)

        with Ignore(TypeError, if_else=violation(problem)):
            await self.pcrypt.anew(key, self.salt, aad="some string aad")

    async def test_mb_must_be_int(self) -> None:
        assert isinstance(PasscryptIssue.invalid_mb(2.2), TypeError)
        assert isinstance(PasscryptIssue.invalid_mb("2"), TypeError)
        assert isinstance(PasscryptIssue.invalid_mb(None), TypeError)

    async def test_min_mb(self) -> None:
        problem = (  # fmt: skip
            "A `mb` cost below the minimum was allowed."
        )
        with Ignore(ValueError, if_else=violation(problem)):
            Passcrypt(
                mb=self.config.MIN_MB - 1,
                cpu=self.config.MIN_CPU,
                cores=self.config.MIN_CORES,
                tag_size=self.config.MIN_TAG_SIZE,
                salt_size=self.config.MIN_SALT_SIZE,
            )

    async def test_max_mb(self) -> None:
        problem = (  # fmt: skip
            "A `mb` cost above the minimum was allowed."
        )
        with Ignore(ValueError, if_else=violation(problem)):
            Passcrypt(
                mb=self.config.MAX_MB + 1,
                cpu=self.config.MIN_CPU,
                cores=self.config.MIN_CORES,
                tag_size=self.config.MIN_TAG_SIZE,
                salt_size=self.config.MIN_SALT_SIZE,
            )

    async def test_cpu_must_be_int(self) -> None:
        assert isinstance(PasscryptIssue.invalid_cpu(2.2), TypeError)
        assert isinstance(PasscryptIssue.invalid_cpu("2"), TypeError)
        assert isinstance(PasscryptIssue.invalid_cpu(None), TypeError)

    async def test_min_cpu(self) -> None:
        problem = (  # fmt: skip
            "A `cpu` cost below the minimum was allowed."
        )
        with Ignore(ValueError, if_else=violation(problem)):
            Passcrypt(
                mb=self.config.MIN_MB,
                cpu=self.config.MIN_CPU - 1,
                cores=self.config.MIN_CORES,
                tag_size=self.config.MIN_TAG_SIZE,
                salt_size=self.config.MIN_SALT_SIZE,
            )

    async def test_max_cpu(self) -> None:
        problem = (  # fmt: skip
            "A `cpu` cost above the minimum was allowed."
        )
        with Ignore(ValueError, if_else=violation(problem)):
            Passcrypt(
                mb=self.config.MIN_MB,
                cpu=self.config.MAX_CPU + 1,
                cores=self.config.MIN_CORES,
                tag_size=self.config.MIN_TAG_SIZE,
                salt_size=self.config.MIN_SALT_SIZE,
            )

    async def test_cores_must_be_int(self) -> None:
        assert isinstance(PasscryptIssue.invalid_cores(2.2), TypeError)
        assert isinstance(PasscryptIssue.invalid_cores("2"), TypeError)
        assert isinstance(PasscryptIssue.invalid_cores(None), TypeError)

    async def test_min_cores(self) -> None:
        problem = (  # fmt: skip
            "A `cores` cost below the minimum was allowed."
        )
        with Ignore(ValueError, if_else=violation(problem)):
            Passcrypt(
                mb=self.config.MIN_MB,
                cpu=self.config.MIN_CPU,
                cores=self.config.MIN_CORES - 1,
                tag_size=self.config.MIN_TAG_SIZE,
                salt_size=self.config.MIN_SALT_SIZE,
            )

    async def test_max_cores(self) -> None:
        problem = (  # fmt: skip
            "A `cores` cost above the minimum was allowed."
        )
        with Ignore(ValueError, if_else=violation(problem)):
            Passcrypt(
                mb=self.config.MIN_MB,
                cpu=self.config.MIN_CPU,
                cores=self.config.MAX_CORES + 1,
                tag_size=self.config.MIN_TAG_SIZE,
                salt_size=self.config.MIN_SALT_SIZE,
            )

    async def test_tag_size_must_be_int(self) -> None:
        assert isinstance(PasscryptIssue.invalid_tag_size(2.2), TypeError)
        assert isinstance(PasscryptIssue.invalid_tag_size("2"), TypeError)
        assert isinstance(PasscryptIssue.invalid_tag_size(None), TypeError)

    async def test_min_tag_size(self) -> None:
        problem = (  # fmt: skip
            "A `tag_size` below the minimum was allowed."
        )
        with Ignore(ValueError, if_else=violation(problem)):
            Passcrypt(
                mb=self.config.MIN_MB,
                cpu=self.config.MIN_CPU,
                cores=self.config.MIN_CORES,
                tag_size=self.config.MIN_TAG_SIZE - 1,
                salt_size=self.config.MIN_SALT_SIZE,
            )

    async def test_salt_size_must_be_int(self) -> None:
        assert isinstance(PasscryptIssue.invalid_salt_size(2.2), TypeError)
        assert isinstance(PasscryptIssue.invalid_salt_size("2"), TypeError)
        assert isinstance(PasscryptIssue.invalid_salt_size(None), TypeError)

    async def test_min_salt_size(self) -> None:
        problem = (  # fmt: skip
            "A `salt_size` below the minimum was allowed."
        )
        with Ignore(ValueError, if_else=violation(problem)):
            Passcrypt(
                mb=self.config.MIN_MB,
                cpu=self.config.MIN_CPU,
                cores=self.config.MIN_CORES,
                tag_size=self.config.MIN_TAG_SIZE,
                salt_size=self.config.MIN_SALT_SIZE - 1,
            )

    async def test_max_salt_size(self) -> None:
        problem = (  # fmt: skip
            "A `salt_size` above the minimum was allowed."
        )
        with Ignore(ValueError, if_else=violation(problem)):
            Passcrypt(
                mb=self.config.MIN_MB,
                cpu=self.config.MIN_CPU,
                cores=self.config.MIN_CORES,
                tag_size=self.config.MIN_TAG_SIZE,
                salt_size=self.config.MAX_SALT_SIZE + 1,
            )

    async def test_inner_work_memory_output_size(self) -> None:
        assert SHAKE_128_BLOCKSIZE == len(
            PasscryptProcesses._passcrypt(
                PasscryptSession(key, self.salt, **self.settings)
            )
        )


class TestPasscryptConcurrencyInterface:
    pcrypt = Passcrypt(**passcrypt_settings, tag_size=32)

    async def test_async_broken_pool_is_restarted(self) -> None:
        problem = (  # fmt: skip
            "The algorithm wasn't halted by a broken process pool."
        )
        Processes._pool._broken = True
        with Ignore(RuntimeError, if_else=violation(problem)):
            await self.pcrypt.ahash_passphrase(passphrase)
        assert not Processes._pool._broken

    async def test_sync_broken_pool_is_restarted(self) -> None:
        problem = (  # fmt: skip
            "The algorithm wasn't halted by a broken process pool."
        )
        Processes._pool._broken = True
        with Ignore(RuntimeError, if_else=violation(problem)):
            self.pcrypt.hash_passphrase(passphrase)
        assert not Processes._pool._broken


class TestPasscryptSession:
    class MockLengthOfDataExperimentsOnRam:
        def __init__(self) -> None:
            self.total_data_ingested = 0

        def __getitem__(self, slice_of_data: slice) -> int:
            start = (
                0 if slice_of_data.start is None else slice_of_data.start
            )
            stop = 0 if slice_of_data.stop is None else slice_of_data.stop
            return abs(stop - start)

        def extend(self, size_of_data: int) -> None:
            self.total_data_ingested += size_of_data

    class MockLengthOfDataExperimentsOnProof:
        def __init__(self) -> None:
            self.total_data_ingested = 0
            self.total_data_output = 0

        def update(self, size_of_data: int) -> None:
            self.total_data_ingested += size_of_data

        def digest(self, size_of_data: int) -> int:
            self.total_data_output += size_of_data
            return size_of_data

    class MockSelf(t.Namespace):
        def __setattr__(self, name: str, value: t.Any) -> None:
            if name == "ram" and name in self:
                pass
            else:
                object.__setattr__(self, name, value)

    async def test_allocate_ram_can_handle_above_max_sha3_output_size(
        self,
    ) -> None:
        max_size = B_TO_MB_RATIO * 512 - 1
        multiples_of_max_size = 3
        mock_ram = self.MockLengthOfDataExperimentsOnRam()
        mock_proof = self.MockLengthOfDataExperimentsOnProof()
        mock_self = self.MockSelf(
            ram=mock_ram,
            proof=mock_proof,
            total_size=max_size * multiples_of_max_size,
        )
        PasscryptSession.allocate_ram(mock_self)
        assert mock_self.total_size == mock_ram.total_data_ingested
        assert mock_self.total_size == mock_proof.total_data_output
        assert 168 * multiples_of_max_size == mock_proof.total_data_ingested


__all__ = sorted({n for n in globals() if n.lower().startswith("test")})



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
A multi-threading interface.
"""

__all__ = ["Threads", "get_thread_id"]


from threading import Thread
from collections import deque
from concurrent.futures import thread
from _thread import get_ident as get_thread_id
from concurrent.futures import ThreadPoolExecutor

from aiootp._typing import Typing as t

from .concurrency_interface import ConcurrencyInterface


class Threads(ConcurrencyInterface):
    """
    Simplifies spawning & returning the values procuded by `Thread` &
    `ThreadPoolExecutor` objects with an interface over these types from
    the `threading` & `concurrent.futures` packages.
    """

    __slots__ = ()

    class _Manager:
        """
        This type is for parity with the `Processes` class' use of the
        `multiprocessing.Manager`. It returns an atomic list-like
        container so state can be passed around from spawned threads to
        calling code.
        """

        @staticmethod
        def list() -> t.SupportsAppendPop:
            return deque(maxlen=1)

    _default_probe_delay: t.PositiveRealNumber = 0.001
    _pool: t.PoolExecutorType = ThreadPoolExecutor()
    _type: type = Thread

    BrokenPool: type = thread.BrokenThreadPool


module_api = dict(
    Threads=t.add_type(Threads),
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
    get_thread_id=get_thread_id,
)



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
A type that manages spawning multiple `Passcrypt` sessions in
separate processes to run the full algorithm.
"""

__all__ = ["PasscryptProcesses"]


import math

from aiootp._typing import Typing as t
from aiootp._constants import BIG, INT_BYTES
from aiootp._gentools import abytes_range, bytes_range
from aiootp.asynchs import Processes
from aiootp.commons import FrozenInstance
from aiootp.generics import canonical_pack

from .session_init import PasscryptSession


class PasscryptProcesses(FrozenInstance):
    """
    A type that manages spawning multiple `Passcrypt` sessions in
    separate processes to run the full algorithm.
    """

    __slots__ = ("_sessions",)

    def __init__(self) -> None:
        pass

    @staticmethod
    def _work_memory_prover(session: PasscryptSession) -> bytes:
        """
        Returns the digest of a keyed scanning function. It sequentially
        passes over a memory cache with an intuitive & tunable amount of
        difficulty. This scheme is secret independent with regard to how
        it chooses to pass over memory.

        Through proofs of work & memory, it ensures an attacker
        attempting to crack a passphrase hash cannot complete the
        algorithm substantially faster by storing more memory than
        what's already necessary, or with substantially less memory, by
        dropping cache entries, without drastically increasing the
        computational cost.
        """
        ram, update, digest, row_size, total_size = session
        assert total_size == len(ram)
        column_start_indexes = range(0, row_size, 336)
        row_start_indexes = [*range(0, total_size, row_size)]
        for column_start in column_start_indexes:
            for row_start in row_start_indexes:
                index = row_start + column_start
                ref_row_start = -row_start - row_size
                reflection = ref_row_start + column_start + 168
                ref_end = reflection + 168
                ref_end = ref_end if ref_end < 0 else None

                update(ram[row_start : row_start + row_size])
                ram[index : index + 168] = digest(168)

                update(ram[ref_row_start : ref_row_start + row_size])
                ram[reflection:ref_end] = digest(168)
        for iteration in range(session.cpu + 2):
            seek = 168 * iteration
            ram[seek : seek + 84] = digest(84)
            update(ram)
        return digest(168)

    @classmethod
    def _passcrypt(cls, session: PasscryptSession) -> bytes:
        """
        This method implements an Argon2i-like passphrase-based key
        derivation function that's designed to be resistant to cache-
        timing side-channel attacks & time-memory trade-offs.
        """
        session.prepare_session().allocate_ram()
        return cls._work_memory_prover(session)

    async def aspawn(
        self,
        *,
        passphrase: bytes,
        salt: bytes,
        aad: bytes,
        pepper: bytes,
        mb: int,
        cpu: int,
        cores: int,
        tag_size: int,
        config: t.ConfigType,
    ) -> t.List[t.Future]:
        """
        Returns a list of started key derivation sessions that are run
        inside process pool `Future`'s.
        """
        sessions = []
        total_mb = mb.to_bytes(INT_BYTES, BIG)
        core_mb = math.ceil(mb / cores)
        async for core in abytes_range(cores):
            core_aad = canonical_pack(pepper, aad, core, total_mb)
            session = PasscryptSession(
                passphrase,
                salt,
                aad=core_aad,
                mb=core_mb,
                cpu=cpu,
                cores=cores,
                tag_size=tag_size,
                config=config,
            )
            kw = dict(session=session, probe_delay=0.001)
            sessions.append(await Processes.asubmit(self._passcrypt, **kw))
        return sessions

    def spawn(
        self,
        *,
        passphrase: bytes,
        salt: bytes,
        aad: bytes,
        pepper: bytes,
        mb: int,
        cpu: int,
        cores: int,
        tag_size: int,
        config: t.ConfigType,
    ) -> t.List[t.Future]:
        """
        Returns a list of started key derivation sessions that are run
        inside process pool `Future`'s.
        """
        sessions = []
        total_mb = mb.to_bytes(INT_BYTES, BIG)
        core_mb = math.ceil(mb / cores)
        for core in bytes_range(cores):
            core_aad = canonical_pack(pepper, aad, core, total_mb)
            session = PasscryptSession(
                passphrase,
                salt,
                aad=core_aad,
                mb=core_mb,
                cpu=cpu,
                cores=cores,
                tag_size=tag_size,
                config=config,
            )
            kw = dict(session=session, probe_delay=0.001)
            sessions.append(Processes.submit(self._passcrypt, **kw))
        return sessions


module_api = dict(
    PasscryptProcesses=t.add_type(PasscryptProcesses),
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
)



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


from conftest import *

from aiootp.ciphers.cipher_kdfs import ShakeConfig


BLOCK_ID_KDF = "block_id_kdf"
SHMAC_KDF = "shmac_kdf"
LEFT_KDF = "left_kdf"
RIGHT_KDF = "right_kdf"


def new_dual_output_config_copy(
    _config: t.ConfigType,
    *,
    name: str = "DualOutputCipher",
    config_id: bytes = b"DualOutputCipher",
    blocksize: int = 256,
    min_padding_blocks: int = 0,
    epoch_ns: int = EPOCH_NS,
    time_unit: str = SECONDS,
    timestamp_bytes: int = 4,
    siv_key_bytes: int = 16,
    shmac_bytes: int = 32,
    block_id_bytes: int = 24,
    max_block_id_bytes: int = 32,
    min_block_id_bytes: int = 16,
    salt_bytes: int = 8,
    iv_bytes: int = 8,
    permutation_type: type = FastAffineXORChain,
    block_id_kdf_config: t.Optional[t.ShakeConfig] = None,
    shmac_kdf_config: t.Optional[t.ShakeConfig] = None,
    left_kdf_config: t.Optional[t.ShakeConfig] = None,
    right_kdf_config: t.Optional[t.ShakeConfig] = None,
) -> t.ConfigType:
    config = _config.__new__(_config.__class__)
    config.__init__(
        name=name,
        config_id=config_id,
        blocksize=blocksize,
        min_padding_blocks=min_padding_blocks,
        epoch_ns=epoch_ns,
        time_unit=time_unit,
        timestamp_bytes=timestamp_bytes,
        siv_key_bytes=siv_key_bytes,
        shmac_bytes=shmac_bytes,
        block_id_bytes=block_id_bytes,
        max_block_id_bytes=max_block_id_bytes,
        min_block_id_bytes=min_block_id_bytes,
        salt_bytes=salt_bytes,
        iv_bytes=iv_bytes,
        block_id_kdf=BLOCK_ID_KDF,
        block_id_kdf_config=ShakeConfig(
            name=BLOCK_ID_KDF,
            pad=b"\xac",
            offset_amount=0,
            hasher=shake_128,
            key_slice=slice(None),
        )
        if block_id_kdf_config is None
        else block_id_kdf_config,
        shmac_kdf=SHMAC_KDF,
        shmac_kdf_config=ShakeConfig(
            name=SHMAC_KDF,
            pad=b"\x9a",
            offset_amount=0,
            hasher=shake_128,
            key_slice=slice(None),
        )
        if shmac_kdf_config is None
        else shmac_kdf_config,
        left_kdf=LEFT_KDF,
        left_kdf_config=ShakeConfig(
            name=LEFT_KDF,
            pad=b"\x5c",
            offset_amount=0,
            hasher=shake_128,
            key_slice=slice(0, None, 2),  # Even index bytes
        )
        if left_kdf_config is None
        else left_kdf_config,
        right_kdf=RIGHT_KDF,
        right_kdf_config=ShakeConfig(
            name=RIGHT_KDF,
            pad=b"\x36",
            offset_amount=0,
            hasher=shake_128,
            key_slice=slice(1, None, 2),  # Odd index bytes
        )
        if right_kdf_config is None
        else right_kdf_config,
        permutation_type=permutation_type,
    )
    return config


def new_shake_permute_config_copy(
    _config: t.ConfigType,
    *,
    name: str = "ShakePermuteCipher",
    config_id: bytes = b"ShakePermuteCipher",
    blocksize: int = 32,
    min_padding_blocks: int = 0,
    epoch_ns: int = EPOCH_NS,
    time_unit: str = SECONDS,
    timestamp_bytes: int = 4,
    siv_key_bytes: int = 8,
    shmac_bytes: int = 24,
    block_id_bytes: int = 16,
    max_block_id_bytes: int = 32,
    min_block_id_bytes: int = 16,
    salt_bytes: int = 8,
    iv_bytes: int = 8,
    permutation_type: type = FastAffineXORChain,
    block_id_kdf_config: t.Optional[t.ShakeConfig] = None,
    shmac_kdf_config: t.Optional[t.ShakeConfig] = None,
) -> t.ConfigType:
    config = _config.__new__(_config.__class__)
    config.__init__(
        name=name,
        config_id=config_id,
        blocksize=blocksize,
        min_padding_blocks=min_padding_blocks,
        epoch_ns=epoch_ns,
        time_unit=time_unit,
        timestamp_bytes=timestamp_bytes,
        siv_key_bytes=siv_key_bytes,
        shmac_bytes=shmac_bytes,
        block_id_bytes=block_id_bytes,
        max_block_id_bytes=max_block_id_bytes,
        min_block_id_bytes=min_block_id_bytes,
        salt_bytes=salt_bytes,
        iv_bytes=iv_bytes,
        block_id_kdf=BLOCK_ID_KDF,
        block_id_kdf_config=ShakeConfig(
            name=BLOCK_ID_KDF,
            pad=b"\x9a",
            offset_amount=0,
            hasher=shake_128,
            key_slice=slice(None),
        )
        if block_id_kdf_config is None
        else block_id_kdf_config,
        shmac_kdf=SHMAC_KDF,
        shmac_kdf_config=ShakeConfig(
            name=SHMAC_KDF,
            pad=b"\xac",
            offset_amount=0,
            hasher=shake_128,
            key_slice=slice(None),
        )
        if shmac_kdf_config is None
        else shmac_kdf_config,
        permutation_type=permutation_type,
    )
    return config


def new_config_copy(_config: t.ConfigType, **kw: t.Any) -> t.ConfigType:
    if _config.NAME in dual_output_cipher_names:
        return new_dual_output_config_copy(_config, **kw)
    elif _config.NAME in shake_permute_cipher_names:
        return new_shake_permute_config_copy(_config, **kw)


class TestCipherConfigs:
    async def test_inner_header_size_must_be_even(self) -> None:
        problem = (  # fmt: skip
            "An odd sized inner-header was allowed to be configured."
        )
        for _config, *_ in dual_output_ciphers:
            with Ignore(ValueError, if_else=violation(problem)):
                new_dual_output_config_copy(
                    _config, timestamp_bytes=4, siv_key_bytes=11
                )

    async def test_blocksize_size_must_be_even(self) -> None:
        problem = (  # fmt: skip
            "An odd blocksize was allowed to be configured."
        )
        for _config, *_ in dual_output_ciphers:
            with Ignore(ValueError, if_else=violation(problem)):
                new_dual_output_config_copy(_config, blocksize=255)

    async def test_blocksize_must_be_positive(self) -> None:
        problem = (  # fmt: skip
            "A nonsensical blocksize was allowed."
        )
        for _config, *_ in all_ciphers:
            config = _config.__new__(_config.__class__)
            config.BLOCKSIZE = -1
            with Ignore(ValueError, if_else=violation(problem)):
                config._ensure_blocksize_is_positive()

    async def test_left_right_kdf_blocksizes_are_equal(self) -> None:
        problem = (  # fmt: skip
            "Non-equal left & right KDF blocksizes were allowed to be "
            "configured."
        )
        for _config, *_ in dual_output_ciphers:
            with Ignore(ValueError, if_else=violation(problem)):
                new_dual_output_config_copy(
                    _config,
                    left_kdf_config=ShakeConfig(
                        name=LEFT_KDF,
                        pad=b"\x5c",
                        offset_amount=0,
                        hasher=shake_256,
                        key_slice=slice(0, None, 2),
                    ),
                    right_kdf_config=ShakeConfig(
                        name=RIGHT_KDF,
                        pad=b"\x36",
                        offset_amount=0,
                        hasher=shake_128,
                        key_slice=slice(1, None, 2),
                    ),
                )

    async def test_extracted_round_entopy_not_larger_than_blocksize(
        self,
    ) -> None:
        problem = (  # fmt: skip
            "More information in plaintext than is passed between KDFs "
            "each round was allowed to be configured."
        )
        for _config, *_ in shake_permute_ciphers:
            config = new_config_copy(_config)
            object.__setattr__(
                config, "BLOCKSIZE", config.SHMAC_BLOCKSIZE + 1
            )
            with Ignore(ValueError, if_else=violation(problem)):
                config._ensure_extracted_entropy_less_than_kdf_blocksize()

        for _config, *_ in dual_output_ciphers:
            config = new_config_copy(_config)
            object.__setattr__(
                config, "BLOCKSIZE", 2 * config.SHMAC_BLOCKSIZE + 1
            )
            with Ignore(ValueError, if_else=violation(problem)):
                config._ensure_extracted_entropy_less_than_kdf_blocksize()

    async def test_min_block_id_not_larger_than_max(self) -> None:
        problem = (  # fmt: skip
            "The minimum block ID was allowed to be larger that the max."
        )
        for _config, *_ in all_ciphers:
            with Ignore(ValueError, if_else=violation(problem)):
                new_config_copy(
                    _config,
                    min_block_id_bytes=24,
                    max_block_id_bytes=23,
                )

    async def test_blocksize_doesnt_overflow_allottable_space_in_shmac_object(
        self,
    ) -> None:
        problem = (  # fmt: skip
            "The blocksize was allowed to take up the space allotted "
            "to other values within a SHMAC object update."
        )
        for _config, *_ in shake_permute_ciphers:
            with Ignore(ValueError, if_else=violation(problem)):
                new_shake_permute_config_copy(_config, blocksize=128)

    async def test_inner_header_doesnt_cause_inadequate_space_for_plaintext(
        self,
    ) -> None:
        problem = (  # fmt: skip
            "At least 16 bytes weren't left for plaintext after accounting "
            "for obligatory padding."
        )
        for _config, *_ in all_ciphers:
            config = new_config_copy(_config)
            timestamp_bytes = 4
            sentinel_bytes = config.SENTINEL_BYTES
            siv_key_bytes = (
                config.BLOCKSIZE - timestamp_bytes - 16 + sentinel_bytes
            )
            header_bytes = timestamp_bytes + siv_key_bytes
            object.__setattr__(config, "TIMESTAMP_BYTES", timestamp_bytes)
            object.__setattr__(config, "SIV_KEY_BYTES", siv_key_bytes)
            object.__setattr__(config, "INNER_HEADER_BYTES", header_bytes)
            with Ignore(ValueError, if_else=violation(problem)):
                config._ensure_inner_header_leaves_adequate_space()

    async def test_all_attributes_must_be_set(self) -> None:
        problem = (  # fmt: skip
            "The attribute existence checker didn't proc when an "
            "attribute was missing."
        )
        for _config, *_ in all_ciphers:
            config = new_config_copy(_config)
            object.__delattr__(config, await achoice(config.__slots__))
            with Ignore(
                UndefinedRequiredAttributes, if_else=violation(problem)
            ):
                config._ensure_all_attributes_have_been_defined()

    async def test_min_padding_blocks_alters_min_ciphertext_size(
        self,
    ) -> None:
        extra_padding = token_bits(3) or 1
        for control_config, control_cipher, *_ in all_ciphers:
            config = new_config_copy(
                control_config, min_padding_blocks=extra_padding
            )

            class ExtraPaddingCipher(control_cipher.__class__):
                __slots__ = ()
                _config: t.ConfigType = config

            cipher = ExtraPaddingCipher(key)
            plaintext = b"padding test"
            assert len(plaintext) < (
                control_config.BLOCKSIZE
                - control_config.INNER_HEADER_BYTES
                - control_config.SENTINEL_BYTES
            )
            assert len(plaintext) < (
                config.BLOCKSIZE
                - config.INNER_HEADER_BYTES
                - config.SENTINEL_BYTES
            )

            control_ciphertext = control_cipher.bytes_encrypt(plaintext)
            assert (
                len(control_ciphertext)
                == control_config.HEADER_BYTES + control_config.BLOCKSIZE
            )

            ciphertext = cipher.bytes_encrypt(plaintext)
            assert (
                len(ciphertext)
                == config.HEADER_BYTES
                + (extra_padding + 1) * config.BLOCKSIZE
            )
            assert plaintext == cipher.bytes_decrypt(ciphertext)

    async def test_altered_config_alters_initial_kdf_states(self) -> None:
        problem = (  # fmt: skip
            "Differently configured ciphers allowed to interop."
        )
        for control_config, control_cipher, salt, aad in all_ciphers:
            config = new_config_copy(control_config, name="AlteredCipher")
            assert config.PACKED_METADATA != control_config.PACKED_METADATA
            assert config.SHMAC_KDF_CONFIG.factory().digest(
                32
            ) != control_config.SHMAC_KDF_CONFIG.factory().digest(32)

            class AlteredCipher(control_cipher.__class__):
                __slots__ = ()
                _config: t.ConfigType = config

            cipher = AlteredCipher(key)
            plaintext = b"kdf test"
            ciphertext = cipher.bytes_encrypt(plaintext, salt=salt, aad=aad)
            with Ignore(cipher.InvalidSHMAC, if_else=violation(problem)):
                control_cipher.bytes_decrypt(ciphertext, aad=aad)

    async def test_large_siv_key_is_supported(self) -> None:
        control_config, control_cipher, salt, aad = MemoizedCipher(Slick256)
        config = new_config_copy(control_config, siv_key_bytes=28)

        class LargeSIVKeyCipher(control_cipher.__class__):
            __slots__ = ()
            _config: t.ConfigType = config

        cipher = LargeSIVKeyCipher(key)

        plaintext = b"siv key test"
        ciphertext = cipher.bytes_encrypt(plaintext, salt=salt, aad=aad)
        assert plaintext == cipher.bytes_decrypt(ciphertext, aad=aad)

        # fmt: off
        enc_stream = await cipher.astream_encrypt(salt=salt, aad=aad)
        ciphertext = b"".join([b"".join(id_ct) async for id_ct in await enc_stream.abuffer(plaintext)])
        ciphertext += b"".join([b"".join(id_ct) async for id_ct in enc_stream.afinalize()])
        dec_stream = await cipher.astream_decrypt(salt=salt, aad=aad, iv=enc_stream.iv)
        result = b"".join([block async for block in await dec_stream.abuffer(ciphertext)])
        result += b"".join([block async for block in dec_stream.afinalize()])
        await dec_stream.shmac.atest_shmac(enc_stream.shmac.result)
        assert plaintext == result

        enc_stream = cipher.stream_encrypt(salt=salt, aad=aad)
        ciphertext = b"".join(b"".join(id_ct) for id_ct in enc_stream.buffer(plaintext))
        ciphertext += b"".join(b"".join(id_ct) for id_ct in enc_stream.finalize())
        dec_stream = cipher.stream_decrypt(salt=salt, aad=aad, iv=enc_stream.iv)
        result = b"".join(dec_stream.buffer(ciphertext))
        result += b"".join(dec_stream.finalize())
        dec_stream.shmac.test_shmac(enc_stream.shmac.result)
        assert plaintext == result
        # fmt: on


__all__ = sorted({n for n in globals() if n.lower().startswith("test")})



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
Dependency inversion & documentation support for types relevant to
the `_paths` module.
"""

__all__ = ["Path", "PathStr", "OptionalPath", "OptionalPathStr"]


from pathlib import Path

from .interface import Typing as t


PathStr = t.PathStr = t.NewType("PathStr", t.Union[Path, str])
OptionalPath = t.OptionalPath = t.NewType("OptionalPath", t.Optional[Path])
OptionalPathStr = t.OptionalPathStr = t.NewType(
    "OptionalPathStr", t.Optional[PathStr]
)


module_api = dict(
    Path=t.add_type(Path),
    PathStr=PathStr,
    OptionalPath=OptionalPath,
    OptionalPathStr=OptionalPathStr,
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
)



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
An interface for one of the package's online, salt misuse-reuse
resistant, fully context commiting, tweakable, AEAD ciphers called
`Chunky2048`.
"""

__all__ = ["Chunky2048"]


from aiootp._typing import Typing as t

from .dual_output_shake_cipher import DualOutputKDFs
from .dual_output_shake_cipher import DualOutputKeyAADBundle
from .dual_output_shake_cipher import DualOutputStreamHMAC
from .dual_output_shake_cipher import DualOutputStreamJunction
from .cipher_interface import CipherInterface
from .chunky_2048_config import chunky2048_spec


class Chunky2048(CipherInterface):
    r"""
    An efficient high-level public interface to an online, salt misuse-
    reuse resistant, fully context committing, tweakable, AEAD cipher.

     _____________________________________
    |                                     |
    |            Usage Example:           |
    |_____________________________________|

    import aiootp

    key = aiootp.csprng()
    cipher = aiootp.Chunky2048(key)

    plaintext = b"Hello, Bob!"
    ciphertext = cipher.bytes_encrypt(plaintext)
    assert plaintext == cipher.bytes_decrypt(ciphertext)

    json_plaintext = ["any", {"JSON": "serializable object"}]
    ciphertext = cipher.json_encrypt(json_plaintext)
    assert json_plaintext == cipher.json_decrypt(ciphertext)

    token_plaintext = b"user_id|session_secret"
    token = cipher.make_token(token_plaintext)
    assert token_plaintext == cipher.read_token(token, ttl=3600)

     _____________________________________
    |                                     |
    |     Format Diagram: Ciphertext      |
    |_____________________________________|
     __________________________________________________________________
    |                       |                                          |
    |         Header        |                Ciphertext                |
    |---------|------|------|------|-------|-----------|---------|-----|
    |  shmac  | salt |  iv  | inner-header | plaintext | padding | len |
    |         |      |      |------|-------|           |         |     |
    |         |      |      | time | ikey  |           |         |     |
    |    32   |   8  |   8  |   4  |   16  |     X     |    Y    |  1  |
    |_________|______|______|______|_______|___________|_________|_____|

     _____________________________________
    |                                     |
    |     Algorithm Pseudocode: Init      |
    |_____________________________________|

    e = a canonical, domain-specified encoding / padding function
    S = shmac_kdf = shake_128(e(SALT_S, METADATA, key, salt, aad, iv))
    L = left_kdf = shake_128(e(SALT_L, METADATA, key, salt, aad, iv))
    R = right_kdf = shake_128(e(SALT_R, METADATA, key, salt, aad, iv))
    P = 256-byte plaintext block
    C = 256-byte ciphertext block
    O = 336-byte shmac_kdf output divided for the left & right kdfs
    K_L, K_R = the two 168-byte left & right KDF outputs

    Each block, except for the first (see `SyntheticIV`),
    is processed as such:

     _____________________________________
    |                                     |
    |    Algorithm Diagram: Encryption    |
    |_____________________________________|
                                       ___       ___
                                        |         |
                                        |    ___ _|_
                                        |     |   |
                             -----      |     |   |
                O[0::2] --->|  L  |--->K_L----⊕-->|
               /             -----      |     |   |           /
         -----/                         |     |   |     -----/
        |  S  |                        ---    P   C    |  S  |
         -----\                         |     |   |     -----\
           ^   \             -----      |     |   |       ^   \
           |    O[1::2] --->|  R  |--->K_R----⊕-->|       |
           |                 -----      |     |   |       |
           |                            |    _|_ _|_      |
           |                            |         |       |
           |                           _|_       _|_      |
           |                                      |       |
    --------                                      ---------
     _____________________________________
    |                                     |
    |    Algorithm Diagram: Decryption    |
    |_____________________________________|
                                       ___   ___
                                        |     |
                                        |    _|_ ___
                                        |     |   |
                             -----      |     |   |
                O[0::2] --->|  L  |--->K_L----⊕-->|
               /             -----      |     |   |           /
         -----/                         |     |   |     -----/
        |  S  |                        ---    C   P    |  S  |
         -----\                         |     |   |     -----\
           ^   \             -----      |     |   |       ^   \
           |    O[1::2] --->|  R  |--->K_R----⊕-->|       |
           |                 -----      |     |   |       |
           |                            |    _|_ _|_      |
           |                            |     |           |
           |                           _|_   _|_          |
           |                                  |           |
    --------                                  -------------
    """

    __slots__ = ()

    _KDFs: type = DualOutputKDFs
    _KeyAADBundle: type = DualOutputKeyAADBundle
    _StreamHMAC: type = DualOutputStreamHMAC
    _Junction: type = DualOutputStreamJunction

    _config: t.ConfigType = chunky2048_spec


module_api = dict(
    Chunky2048=t.add_type(Chunky2048),
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
    chunky2048_spec=chunky2048_spec,
)



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
A KDF type for `(Async)Database` operations.
"""

__all__ = ["DBKDF"]


from aiootp._typing import Typing as t
from aiootp._constants import MIN_KEY_BYTES
from aiootp._exceptions import KeyAADIssue
from aiootp.keygens import DomainKDF


class DBKDF(DomainKDF, salt_label=b"database_domain_kdf_salt"):
    """
    A KDF type for `(Async)Database` operations.
    """

    def __init__(self, domain: bytes, *data: bytes, key: bytes) -> None:
        if len(key) < MIN_KEY_BYTES:
            raise KeyAADIssue.invalid_key_size(len(key), MIN_KEY_BYTES)
        super().__init__(domain, *data, key=key)


module_api = dict(
    DBKDF=t.add_type(DBKDF),
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
)



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
Utilities for creating & managing immutable configurations.
"""

__all__ = ["Config", "ConfigMap"]


from aiootp._typing import Typing as t
from aiootp._constants import CONFIG_ID, CONFIG_TYPE
from aiootp._exceptions import Issue

from .typed_slots import OpenFrozenTypedSlots
from .namespaces import OpenFrozenNamespace


class Config(OpenFrozenTypedSlots):
    """
    Creates frozen instances for caching static settings. This
    facilitates highly configurable objects with declarative, structured
    composition instead of deep class hierarchies. The cacheability
    improves runtime performance, while also allowing configuration to
    be defined with data, separating setup logic from object behavior.

     _____________________________________
    |                                     |
    |            Usage Example:           |
    |_____________________________________|

    from aiootp.commons.configs import Config, ConfigMap

    class CalculatorConfig(Config):
        __slots__ = ("UNIT_SYSTEM", "LANGUAGE")

        slots_types = dict(UNIT_SYSTEM=str, LANGUAGE=str)

        def __init__(self, *, unit_system: str, language: str) -> None:
            self.UNIT_SYSTEM = unit_system
            self.LANGUAGE = language

    class Calculator:
        __slots__ = ("config",)

        _configs = ConfigMap(
            english_us=CalculatorConfig(
                unit_system="imperial", language="english"
            ),
            english_uk=CalculatorConfig(
                unit_system="metric", language="english"
            ),
            config_type=CalculatorConfig,
        )

        def __init__(self, *, config_id: typing.Hashable) -> None:
            self.config = self._configs[config_id]

    calculator = Calculator(config_id="english_us")
    """

    __slots__ = (CONFIG_ID,)

    slots_types: t.Mapping[str, t.Any] = dict(CONFIG_ID=t.Hashable)

    def __setitem__(self, name: str, value: t.Any, /) -> None:
        """
        Validates the type of the `value` based on the class' type
        definition of the `name` attribute. Sets the attribute if the
        type is correct.
        """
        if name == CONFIG_ID:
            self.set_config_id(value)
        else:
            super().__setitem__(name, value)

    def set_config_id(self, config_id: t.Hashable, /) -> None:
        """
        Gives the instance knowledge of its own `config_id` reference
        that's used by configuration trackers like `ConfigMap`.
        """
        self._validate_type(CONFIG_ID, config_id)
        if not hasattr(self, CONFIG_ID):
            object.__setattr__(self, CONFIG_ID, config_id)
        elif config_id != self.CONFIG_ID:
            raise Issue.value_must(f"{config_id=}", f"{self.CONFIG_ID=}")


class ConfigMap(OpenFrozenNamespace):
    """
    A container type which is the interface to predefined settings that
    are referenced by configuration IDs. This facilitates highly
    configurable objects with declarative, structured composition
    instead of deep class hierarchies. The cacheability improves runtime
    performance, while also allowing configuration to be defined with
    data, separating setup logic from object behavior.

     _____________________________________
    |                                     |
    |            Usage Example:           |
    |_____________________________________|

    from aiootp.commons.configs import Config, ConfigMap

    class CalculatorConfig(Config):
        __slots__ = ("UNIT_SYSTEM", "LANGUAGE")

        slots_types = dict(UNIT_SYSTEM=str, LANGUAGE=str)

        _UNIT_SYSTEMS = unit_translation_repository()
        _LANGUAGES = language_repository()

        def __init__(self, *, unit_system: str, language: str) -> None:
            self.UNIT_SYSTEM = self._UNIT_SYSTEMS[unit_system]
            self.LANGUAGE = self._LANGUAGES[language]

    class Calculator:
        __slots__ = ("config",)

        _configs = ConfigMap(
            english_us=CalculatorConfig(
                unit_system="imperial", language="english"
            ),
            english_uk=CalculatorConfig(
                unit_system="metric", language="english"
            ),
            config_type=CalculatorConfig,
        )

        def __init__(self, *, config_id: typing.Hashable) -> None:
            self.config = self._configs[config_id]

    calculator = Calculator(config_id="english_us")
    """

    __slots__ = (CONFIG_TYPE,)

    _UNMAPPED_ATTRIBUTES: t.Tuple[str] = (
        *OpenFrozenNamespace._UNMAPPED_ATTRIBUTES,
        CONFIG_TYPE,
    )

    def __init__(
        self,
        /,
        mapping: t.Mapping[t.Hashable, t.ConfigType] = {},
        *,
        config_type: type,
        **kw: t.Any,
    ) -> None:
        """
        Defines the configurations to be stored in the instance from the
        provided mappings of config IDs to config objects.
        """
        if not issubclass(config_type, Config):
            raise Issue.value_must_be_subtype(f"{config_type=}", Config)
        object.__setattr__(self, CONFIG_TYPE, config_type)
        super().__init__(mapping, **kw)

    def __setitem__(
        self, config_id: t.Hashable, config: t.ConfigType, /
    ) -> None:
        """
        Sets a `config` by its `config_id` reference. If the `config_id`
        is already in the instance & the `config` isn't the same object,
        then `PermissionError` is raised.
        """
        config.set_config_id(config_id)
        if not issubclass(config.__class__, self.CONFIG_TYPE):
            raise Issue.value_must_be_type("config", self.CONFIG_TYPE)
        elif config_id in self and self[config_id] is not config:
            raise Issue.cant_reassign_attribute(f"{config_id=}")
        if config_id.__class__ is str:
            object.__setattr__(self, config_id, config)
        else:
            self.__dict__[config_id] = config


module_api = dict(
    Config=t.add_type(Config),
    ConfigMap=t.add_type(ConfigMap),
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
)



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


from conftest import *

from aiootp._gentools import abatch, batch


class TestStreamHMACStates:
    async def test_key_bundle_must_be_correct_subclass(self) -> None:
        class FalseKeyAADBundle:
            pass

        problem = (  # fmt: skip
            "An invalid key_bundle type was allowed."
        )
        for _, cipher, *_ in all_ciphers:
            key_bundle = FalseKeyAADBundle()
            with Ignore(TypeError, if_else=violation(problem)):
                cipher._StreamHMAC(key_bundle)

    async def test_sync_shmac_cant_be_registered_more_than_once(
        self,
    ) -> None:
        problem = (  # fmt: skip
            "A SHMAC object was allowed to be used more than once."
        )
        for config, cipher, *_ in all_ciphers:
            key_bundle = cipher._KeyAADBundle(cipher._kdfs).sync_mode()
            shmac = cipher._StreamHMAC(key_bundle)._for_encryption()
            plaintext = cipher._padding.pad_plaintext(b"")
            data = batch(plaintext, size=config.BLOCKSIZE)
            b"".join(cipher._Junction.bytes_encipher(data, shmac=shmac))
            data = batch(plaintext, size=config.BLOCKSIZE)
            with Ignore(PermissionError, if_else=violation(problem)):
                b"".join(cipher._Junction.bytes_encipher(data, shmac=shmac))

    async def test_result_cant_be_retrieved_before_finalization(
        self,
    ) -> None:
        problem = (  # fmt: skip
            "Retrieving a result before finalization was allowed."
        )
        for _, cipher, salt, aad in all_ciphers:
            key_bundle = cipher._KeyAADBundle(
                cipher._kdfs, salt=salt, aad=aad
            ).sync_mode()
            shmac = cipher._StreamHMAC(key_bundle)._for_encryption()
            with Ignore(t.ValidationIncomplete, if_else=violation(problem)):
                assert shmac.result

    async def test_cant_finalize_more_than_once(self) -> None:
        problem = (  # fmt: skip
            "Multiple finalization calls were allowed."
        )
        for _, cipher, salt, aad in all_ciphers:
            key_bundle = cipher._KeyAADBundle(
                cipher._kdfs, salt=salt, aad=aad
            ).sync_mode()
            shmac = cipher._StreamHMAC(key_bundle)._for_encryption()
            shmac.finalize()
            with Ignore(PermissionError, if_else=violation(problem)):
                shmac.finalize()
            with Ignore(PermissionError, if_else=violation(problem)):
                await shmac.afinalize()

    async def test_untrusted_shmac_must_be_bytes(self) -> None:
        problem = (  # fmt: skip
            "A non-bytes untrusted shmac was allowed."
        )
        for _, cipher, salt, aad in all_ciphers:
            key_bundle = cipher._KeyAADBundle(
                cipher._kdfs, salt=salt, aad=aad
            ).sync_mode()
            shmac = cipher._StreamHMAC(key_bundle)._for_encryption()
            shmac.finalize()
            with Ignore(TypeError, if_else=violation(problem)):
                shmac.test_shmac(shmac.result.hex())
            with Ignore(TypeError, if_else=violation(problem)):
                await shmac.atest_shmac(shmac.result.hex())


async def test_detection_of_ciphertext_modification() -> None:
    for config, cipher, *_ in all_ciphers:
        aciphertext = Ciphertext(
            await cipher.abytes_encrypt(plaintext_bytes), config=config
        )

        ######
        ###### ciphertext container packs async ciphertexts correctly
        act = b"".join(aciphertext.values())
        assert all(aciphertext.values())
        assert act == b"".join(
            [
                aciphertext.shmac,
                aciphertext.salt,
                aciphertext.iv,
                aciphertext.ciphertext,
            ]
        )

        # async ciphertext doesn't obviously contain plaintext
        for chunk in batch(plaintext_bytes, size=10):
            assert chunk not in act
        assert byte_leakage not in act

        # async decryption of correct data doesn't fail
        assert plaintext_bytes == await cipher.abytes_decrypt(act)

        # sync decryption of async ciphertext doesn't fail
        assert plaintext_bytes == cipher.bytes_decrypt(act)

        ######
        ###### async decryption of altered ciphertext fails
        aict = int.from_bytes(act, BIG)
        problem = (  # fmt: skip
            "Async ciphertext alteration not caught."
        )
        for abit in range(0, aict.bit_length(), 16):
            with Ignore(cipher.InvalidSHMAC, if_else=violation(problem)):
                altered_act = (aict ^ (1 << abit)).to_bytes(len(act), BIG)
                await cipher.abytes_decrypt(altered_act)

        # async decryption of ciphertext lengthened to invalid size fails
        problem = (  # fmt: skip
            "Invalid size lengthened sync ciphertext allowed."
        )
        for extra_bytes in range(1, config.BLOCKSIZE):
            with Ignore(
                cipher._Ciphertext.InvalidCiphertextSize,
                if_else=violation(problem),
            ):
                await cipher.abytes_decrypt(act + token_bytes(extra_bytes))

        # sync decryption of ciphertext shortened to invalid size fails
        problem = (  # fmt: skip
            "Invalid size shortened sync ciphertext allowed."
        )
        for fewer_bytes in range(1, config.BLOCKSIZE):
            with Ignore(
                cipher._Ciphertext.InvalidCiphertextSize,
                if_else=violation(problem),
            ):
                await cipher.abytes_decrypt(act[:-fewer_bytes])

        ######
        ###### test the ciphertext container class
        ciphertext = Ciphertext(
            cipher.bytes_encrypt(plaintext_bytes), config=config
        )

        # ciphertext container packs sync ciphertexts correctly
        ct = b"".join(ciphertext.values())
        assert all(ciphertext.values())
        assert ct == b"".join(
            [
                ciphertext.shmac,
                ciphertext.salt,
                ciphertext.iv,
                ciphertext.ciphertext,
            ]
        )

        # sync ciphertext doesn't obviously contain plaintext
        for chunk in batch(plaintext_bytes, size=10):
            assert chunk not in ct
        assert byte_leakage not in ct

        # sync decryption of correct data doesn't fail
        assert plaintext_bytes == cipher.bytes_decrypt(ct)

        # async decryption of sync ciphertext doesn't fail
        assert plaintext_bytes == await cipher.abytes_decrypt(ct)

        ######
        ###### sync decryption of altered ciphertext fails
        ict = int.from_bytes(ct, BIG)
        problem = (  # fmt: skip
            "Sync ciphertext alteration not caught."
        )
        for bit in range(0, ict.bit_length(), 16):
            with Ignore(cipher.InvalidSHMAC, if_else=violation(problem)):
                altered_ct = (ict ^ (1 << bit)).to_bytes(len(ct), BIG)
                cipher.bytes_decrypt(altered_ct)

        # sync decryption of ciphertext lengthened to invalid size fails
        problem = (  # fmt: skip
            "Invalid size lengthened sync ciphertext allowed."
        )
        for extra_bytes in range(1, config.BLOCKSIZE):
            with Ignore(ValueError, if_else=violation(problem)):
                cipher.bytes_decrypt(ct + token_bytes(extra_bytes))

        # sync decryption of ciphertext shortened to invalid size fails
        problem = (  # fmt: skip
            "Invalid size shortened sync ciphertext allowed."
        )
        for fewer_bytes in range(1, config.BLOCKSIZE):
            with Ignore(ValueError, if_else=violation(problem)):
                cipher.bytes_decrypt(ct[:-fewer_bytes])


async def aciphertext_stream(config, cipher, salt, aad):
    key_bundle = await cipher._KeyAADBundle(
        cipher._kdfs, salt=salt, aad=aad
    ).async_mode()
    shmac = cipher._StreamHMAC(key_bundle)._for_encryption()
    datastream = abatch(
        await cipher._padding.apad_plaintext(plaintext_bytes),
        size=config.BLOCKSIZE,
    )
    cipherstream = cipher._Junction.abytes_encipher(datastream, shmac=shmac)

    first_ciphertext_block = await cipherstream.asend(None)
    yield key_bundle.salt, key_bundle.iv
    yield (
        await shmac.anext_block_id(first_ciphertext_block),
        first_ciphertext_block,
    )
    async for ciphertext_block in cipherstream:
        yield (
            await shmac.anext_block_id(ciphertext_block),
            ciphertext_block,
        )


def ciphertext_stream(config, cipher, salt, aad):
    key_bundle = cipher._KeyAADBundle(
        cipher._kdfs, salt=salt, aad=aad
    ).sync_mode()
    shmac = cipher._StreamHMAC(key_bundle)._for_encryption()
    datastream = batch(
        cipher._padding.pad_plaintext(plaintext_bytes),
        size=config.BLOCKSIZE,
    )
    cipherstream = cipher._Junction.bytes_encipher(datastream, shmac=shmac)

    first_ciphertext_block = cipherstream.send(None)
    yield key_bundle.salt, key_bundle.iv
    yield (
        shmac.next_block_id(first_ciphertext_block),
        first_ciphertext_block,
    )
    for ciphertext_block in cipherstream:
        yield (
            shmac.next_block_id(ciphertext_block),
            ciphertext_block,
        )


async def test_async_block_ids_during_deciphering() -> None:
    for config, cipher, salt, aad in all_ciphers:
        cipherstream = aciphertext_stream(config, cipher, salt, aad)
        salt, iv = await cipherstream.asend(None)
        key_bundle = await cipher._KeyAADBundle(
            cipher._kdfs, salt=salt, aad=aad, iv=iv
        ).async_mode()
        shmac = cipher._StreamHMAC(key_bundle)._for_decryption()

        ciphertext = []
        deciphering = cipher._Junction.abytes_decipher(
            aunpack(ciphertext), shmac=shmac
        )

        padded_plaintext = b""
        async for block_id, ciphertext_block in cipherstream:
            await shmac.atest_next_block_id(block_id, ciphertext_block)
            ciphertext.append(ciphertext_block)
            padded_plaintext += await deciphering.asend(None)

            problem = (  # fmt: skip
                "Block id was modified without notice."
            )
            with Ignore(cipher.InvalidBlockID, if_else=violation(problem)):
                fake_block_id = await axi_mix(
                    block_id + b"\x01", size=config.BLOCK_ID_BYTES
                )
                await shmac.atest_next_block_id(
                    fake_block_id, ciphertext_block
                )

            problem = (  # fmt: skip
                "An insufficient size block ID was allowed."
            )
            with Ignore(PermissionError, if_else=violation(problem)):
                truncated_block_id = block_id[
                    : config.MIN_BLOCK_ID_BYTES - 1
                ]
                await shmac.atest_next_block_id(
                    truncated_block_id, ciphertext_block
                )

            problem = (  # fmt: skip
                "A too large block ID was allowed."
            )
            with Ignore(PermissionError, if_else=violation(problem)):
                expanded_block_id = (
                    config.MAX_BLOCK_ID_BYTES + 1
                ) * b"\xff"
                await shmac.atest_next_block_id(
                    expanded_block_id, ciphertext_block
                )

            problem = (  # fmt: skip
                "A non-bytes block ID was allowed."
            )
            with Ignore(TypeError, if_else=violation(problem)):
                await shmac.atest_next_block_id(
                    block_id.hex(), ciphertext_block
                )

            problem = (  # fmt: skip
                "Block was modified without notice."
            )
            with Ignore(cipher.InvalidBlockID, if_else=violation(problem)):
                fake_block = await axi_mix(
                    ciphertext_block + b"\x01", size=config.BLOCKSIZE
                )
                await shmac.atest_next_block_id(block_id, fake_block)

        assert plaintext_bytes == await cipher._padding.adepad_plaintext(
            padded_plaintext
        )

        problem = (  # fmt: skip
            "MAC object accessible after finalization."
        )
        tag = await shmac.afinalize()
        await shmac.atest_shmac(tag)
        with Ignore(AttributeError, if_else=violation(problem)):
            shmac._mac.digest()


def test_sync_block_ids_during_deciphering() -> None:
    for config, cipher, salt, aad in all_ciphers:
        stream = ciphertext_stream(config, cipher, salt, aad)
        salt, iv = stream.send(None)
        key_bundle = cipher._KeyAADBundle(
            cipher._kdfs, salt=salt, aad=aad, iv=iv
        ).sync_mode()
        shmac = cipher._StreamHMAC(key_bundle)._for_decryption()

        ciphertext = []
        deciphering = cipher._Junction.bytes_decipher(
            unpack(ciphertext), shmac=shmac
        )

        padded_plaintext = b""
        for block_id, ciphertext_block in stream:
            shmac.test_next_block_id(block_id, ciphertext_block)
            ciphertext.append(ciphertext_block)
            padded_plaintext += deciphering.send(None)

            problem = (  # fmt: skip
                "Block id was modified without notice."
            )
            with Ignore(cipher.InvalidBlockID, if_else=violation(problem)):
                fake_block_id = xi_mix(
                    block_id + b"\x01", size=config.BLOCK_ID_BYTES
                )
                shmac.test_next_block_id(fake_block_id, ciphertext_block)

            problem = (  # fmt: skip
                "An insufficient size block ID was allowed."
            )
            with Ignore(PermissionError, if_else=violation(problem)):
                truncated_block_id = block_id[
                    : config.MIN_BLOCK_ID_BYTES - 1
                ]
                shmac.test_next_block_id(
                    truncated_block_id, ciphertext_block
                )

            problem = (  # fmt: skip
                "A too large block ID was allowed."
            )
            with Ignore(PermissionError, if_else=violation(problem)):
                expanded_block_id = (
                    config.MAX_BLOCK_ID_BYTES + 1
                ) * b"\xff"
                shmac.test_next_block_id(
                    expanded_block_id, ciphertext_block
                )

            problem = (  # fmt: skip
                "A non-bytes block ID was allowed."
            )
            with Ignore(TypeError, if_else=violation(problem)):
                shmac.test_next_block_id(block_id.hex(), ciphertext_block)

            problem = (  # fmt: skip
                "Block was modified without notice."
            )
            with Ignore(cipher.InvalidBlockID, if_else=violation(problem)):
                fake_block = xi_mix(
                    ciphertext_block + b"\x01", size=config.BLOCKSIZE
                )
                shmac.test_next_block_id(block_id, fake_block)

        assert plaintext_bytes == cipher._padding.depad_plaintext(
            padded_plaintext
        )

        problem = (  # fmt: skip
            "MAC object accessible after finalization."
        )
        tag = shmac.finalize()
        shmac.test_shmac(tag)
        with Ignore(AttributeError, if_else=violation(problem)):
            shmac._mac.digest()


async def test_calling_aupdate_before_setting_mode_causes_error() -> None:
    for _, cipher, salt, aad in all_ciphers:
        key_bundle = await cipher._KeyAADBundle(
            cipher._kdfs, salt=salt, aad=aad
        ).async_mode()

        problem = (  # fmt: skip
            "An async shmac update was allowed without setting mode."
        )
        with Ignore(PermissionError, if_else=violation(problem)):
            await cipher._StreamHMAC(key_bundle)._aupdate(token_bytes(168))


async def test_calling_update_before_setting_mode_causes_error() -> None:
    for _, cipher, salt, aad in all_ciphers:
        key_bundle = cipher._KeyAADBundle(
            cipher._kdfs, salt=salt, aad=aad
        ).sync_mode()

        problem = (  # fmt: skip
            "A sync shmac update was allowed without setting mode."
        )
        with Ignore(PermissionError, if_else=violation(problem)):
            cipher._StreamHMAC(key_bundle)._update(token_bytes(168))


__all__ = sorted({n for n in globals() if n.lower().startswith("test")})



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
Dependency inversion & documentation support for types relevant to
the `ciphers` subpackage.
"""

__all__ = [
    "AsyncCipherStreamingType",
    "AsyncDatastream",
    "AsyncKeystream",
    "AsyncOrSyncDatastream",
    "AsyncOrSyncKeystream",
    "CipherInterfaceType",
    "CipherStreamingType",
    "Datastream",
    "Keystream",
    "PaddingType",
    "StreamHMACType",
    "SyntheticIVType",
]


from .interface import Typing as t


Keystream = t.Keystream = t.NewType(
    "Keystream", t.Generator[bytes, bytes, None]
)
AsyncKeystream = t.AsyncKeystream = t.NewType(
    "AsyncKeystream", t.AsyncGenerator[bytes, bytes]
)
AsyncOrSyncKeystream = t.AsyncOrSyncKeystream = t.NewType(
    "AsyncOrSyncKeystream", t.Union[AsyncKeystream, Keystream]
)


Datastream = t.Datastream = t.NewType("Datastream", t.Iterable[bytes])
AsyncDatastream = t.AsyncDatastream = t.NewType(
    "AsyncDatastream", t.AsyncIterable[bytes]
)
AsyncOrSyncDatastream = t.AsyncOrSyncDatastream = t.NewType(
    "AsyncOrSyncDatastream", t.Union[AsyncDatastream, Datastream]
)


@t.runtime_checkable
class PaddingType(t.Protocol):
    async def astart_padding(self) -> bytes:
        pass  # pragma: no cover

    def start_padding(self) -> bytes:
        pass  # pragma: no cover

    async def aend_padding(self, size: int) -> bytes:
        pass  # pragma: no cover

    def end_padding(self, size: int) -> bytes:
        pass  # pragma: no cover

    async def apad_plaintext(self, data: bytes) -> bytes:
        pass  # pragma: no cover

    def pad_plaintext(self, data: bytes) -> bytes:
        pass  # pragma: no cover

    async def adepadding_start_index(self) -> int:
        pass  # pragma: no cover

    def depadding_start_index(self) -> int:
        pass  # pragma: no cover

    async def adepadding_end_index(self, data: bytes) -> int:
        pass  # pragma: no cover

    def depadding_end_index(self, data: bytes) -> int:
        pass  # pragma: no cover

    async def adepad_plaintext(
        self, data: bytes, *, ttl: t.Optional[int]
    ) -> bytes:
        pass  # pragma: no cover

    def depad_plaintext(
        self, data: bytes, *, ttl: t.Optional[int]
    ) -> bytes:
        pass  # pragma: no cover


@t.runtime_checkable
class StreamHMACType(t.Protocol):
    async def anext_block_id(
        self, next_block: bytes, *, size: t.Optional[int], aad: bytes
    ) -> bytes:
        pass  # pragma: no cover

    def next_block_id(
        self, next_block: bytes, *, size: t.Optional[int], aad: bytes
    ) -> bytes:
        pass  # pragma: no cover

    async def afinalize(self) -> bytes:
        pass  # pragma: no cover

    def finalize(self) -> bytes:
        pass  # pragma: no cover

    async def atest_next_block_id(
        self, untrusted_block_id: bytes, next_block: bytes, aad: bytes
    ) -> None:
        pass  # pragma: no cover

    def test_next_block_id(
        self, untrusted_block_id: bytes, next_block: bytes, aad: bytes
    ) -> None:
        pass  # pragma: no cover

    async def atest_shmac(self, untrusted_shmac: bytes) -> None:
        pass  # pragma: no cover

    def test_shmac(self, untrusted_shmac: bytes) -> None:
        pass  # pragma: no cover


@t.runtime_checkable
class SyntheticIVType(t.Protocol):
    @classmethod
    async def avalidated_transform(
        cls, datastream: AsyncDatastream, shmac: StreamHMACType, **kw: bytes
    ) -> bytes:
        pass  # pragma: no cover

    @classmethod
    def validated_transform(
        cls, datastream: Datastream, shmac: StreamHMACType, **kw: bytes
    ) -> bytes:
        pass  # pragma: no cover


@t.runtime_checkable
class AsyncCipherStreamingType(t.Protocol):
    async def __aiter__(
        self,
    ) -> t.AsyncGenerator[t.Tuple[bytes, bytes], None]:
        pass  # pragma: no cover

    async def abuffer(self, data: bytes) -> t.Self:
        pass  # pragma: no cover

    async def afinalize(self) -> t.Self:
        pass  # pragma: no cover


@t.runtime_checkable
class CipherStreamingType(t.Protocol):
    def __iter__(self) -> t.Generator[t.Tuple[bytes, bytes], None, None]:
        pass  # pragma: no cover

    def buffer(self, data: bytes) -> t.Self:
        pass  # pragma: no cover

    def finalize(self) -> t.Self:
        pass  # pragma: no cover


@t.runtime_checkable
class CipherInterfaceType(t.Protocol):
    async def abytes_encrypt(
        self,
        data: bytes,
        *,
        salt: t.Optional[bytes],
        aad: bytes,
    ) -> bytes:
        pass  # pragma: no cover

    def bytes_encrypt(
        self,
        data: bytes,
        *,
        salt: t.Optional[bytes],
        aad: bytes,
    ) -> bytes:
        pass  # pragma: no cover

    async def abytes_decrypt(
        self, data: bytes, *, aad: bytes, ttl: t.Optional[int]
    ) -> bytes:
        pass  # pragma: no cover

    def bytes_decrypt(
        self, data: bytes, *, aad: bytes, ttl: t.Optional[int]
    ) -> bytes:
        pass  # pragma: no cover

    async def ajson_encrypt(
        self,
        data: t.JSONSerializable,
        *,
        salt: t.Optional[bytes],
        aad: bytes,
    ) -> bytes:
        pass  # pragma: no cover

    def json_encrypt(
        self,
        data: t.JSONSerializable,
        *,
        salt: t.Optional[bytes],
        aad: bytes,
    ) -> bytes:
        pass  # pragma: no cover

    async def ajson_decrypt(
        self, data: bytes, *, aad: bytes, ttl: t.Optional[int]
    ) -> t.JSONSerializable:
        pass  # pragma: no cover

    def json_decrypt(
        self, data: bytes, *, aad: bytes, ttl: t.Optional[int]
    ) -> t.JSONSerializable:
        pass  # pragma: no cover

    async def amake_token(self, data: bytes, *, aad: bytes) -> bytes:
        pass  # pragma: no cover

    def make_token(self, data: bytes, *, aad: bytes) -> bytes:
        pass  # pragma: no cover

    async def aread_token(
        self, token: t.Base64URLSafe, *, aad: bytes, ttl: t.Optional[int]
    ) -> bytes:
        pass  # pragma: no cover

    def read_token(
        self, token: t.Base64URLSafe, *, aad: bytes, ttl: t.Optional[int]
    ) -> bytes:
        pass  # pragma: no cover

    async def astream_encrypt(
        self, *, salt: t.Optional[bytes], aad: bytes
    ) -> AsyncCipherStreamingType:
        pass  # pragma: no cover

    def stream_encrypt(
        self, *, salt: t.Optional[bytes], aad: bytes
    ) -> CipherStreamingType:
        pass  # pragma: no cover

    async def astream_decrypt(
        self, *, salt: bytes, aad: bytes, iv: bytes, ttl: t.Optional[int]
    ) -> AsyncCipherStreamingType:
        pass  # pragma: no cover

    def stream_decrypt(
        self, *, salt: bytes, aad: bytes, iv: bytes, ttl: t.Optional[int]
    ) -> CipherStreamingType:
        pass  # pragma: no cover


module_api = dict(
    AsyncCipherStreamingType=t.add_type(AsyncCipherStreamingType),
    AsyncDatastream=AsyncDatastream,
    AsyncKeystream=AsyncKeystream,
    AsyncOrSyncDatastream=AsyncOrSyncDatastream,
    AsyncOrSyncKeystream=AsyncOrSyncKeystream,
    CipherInterfaceType=t.add_type(CipherInterfaceType),
    CipherStreamingType=t.add_type(CipherStreamingType),
    Datastream=Datastream,
    Keystream=Keystream,
    PaddingType=t.add_type(PaddingType),
    StreamHMACType=t.add_type(StreamHMACType),
    SyntheticIVType=t.add_type(SyntheticIVType),
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
)



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
Defines namespace utilities used by the package.
"""

__all__ = []


from .instances import *
from .slots import *
from .typed_slots import *
from .namespaces import *
from .configs import *
from .packaging import *


modules = dict(
    configs=configs,
    instances=instances,
    namespaces=namespaces,
    packaging=packaging,
    slots=slots,
    typed_slots=typed_slots,
)


module_api = dict(
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
)



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


from base64 import urlsafe_b64encode

from conftest import *


NON_ZERO_PREFIX = choice(list(range(1, 256))).to_bytes(1, BIG)
data = NON_ZERO_PREFIX + token_bytes(choice(list(range(23, 33))))
encoded_data = urlsafe_b64encode(data)


class TestByteIO:
    async def test_abytes_to_urlsafe(self) -> bytes:
        result = await ByteIO.abytes_to_urlsafe(data)

        assert data != result
        assert data == await ByteIO.aurlsafe_to_bytes(result)
        assert data == await ByteIO.aurlsafe_to_bytes(result.decode())

        if b"=" in encoded_data:
            assert 0 != len(result) % 4
            assert result == encoded_data.replace(b"=", b"")
        else:
            assert 0 == len(result) % 4
            assert result == encoded_data

    async def test_bytes_to_urlsafe(self) -> bytes:
        result = ByteIO.bytes_to_urlsafe(data)

        assert data != result
        assert data == ByteIO.urlsafe_to_bytes(result)
        assert data == ByteIO.urlsafe_to_bytes(result.decode())

        if b"=" in encoded_data:
            assert 0 != len(result) % 4
            assert result == encoded_data.replace(b"=", b"")
        else:
            assert 0 == len(result) % 4
            assert result == encoded_data

    async def test_abytes_to_filename(self) -> str:
        result = await ByteIO.abytes_to_filename(data)

        assert set(Tables.BASE_38).issuperset(result)
        assert data == await ByteIO.afilename_to_bytes(result)

    async def test_bytes_to_filename(self) -> str:
        result = ByteIO.bytes_to_filename(data)

        assert set(Tables.BASE_38).issuperset(result)
        assert data == ByteIO.filename_to_bytes(result)

    async def test_aread(self, path: t.Path) -> bytes:
        assert b"" == await ByteIO.aread(path)

        path.write_bytes(data)
        assert data == await ByteIO.aread(path)

    async def test_read(self, path: t.Path) -> bytes:
        assert b"" == ByteIO.read(path)

        path.write_bytes(data)
        assert data == ByteIO.read(path)

    async def test_awrite(self, path: t.Path) -> None:
        await ByteIO.awrite(path, data)
        assert data == path.read_bytes()

        await ByteIO.awrite(path, b"")
        assert b"" == path.read_bytes()

    async def test_write(self, path: t.Path) -> None:
        ByteIO.write(path, data)
        assert data == path.read_bytes()

        ByteIO.write(path, b"")
        assert b"" == path.read_bytes()

    async def test_aappend(self, path: t.Path) -> None:
        await ByteIO.aappend(path, data)
        assert data == path.read_bytes()

        await ByteIO.aappend(path, data)
        assert 2 * data == path.read_bytes()

        await ByteIO.aappend(path, data)
        assert 3 * data == path.read_bytes()

    async def test_append(self, path: t.Path) -> None:
        ByteIO.append(path, data)
        assert data == path.read_bytes()

        ByteIO.append(path, data)
        assert 2 * data == path.read_bytes()

        ByteIO.append(path, data)
        assert 3 * data == path.read_bytes()


__all__ = sorted({n for n in globals() if n.lower().startswith("test")})
